package jp.moke;

import java.util.Iterator;
import java.util.Set;

import android.content.SharedPreferences;
import android.content.SharedPreferences.Editor;
import android.os.Bundle;
import android.preference.EditTextPreference;
import android.preference.ListPreference;
import android.preference.Preference;
import android.preference.Preference.OnPreferenceChangeListener;
import android.preference.PreferenceActivity;
import android.preference.PreferenceManager;
import android.widget.Toast;

public class BookmarkPrefActivity extends PreferenceActivity {

	SharedPreferences defaultSharedPreferences;

	@Override
	protected void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		addPreferencesFromResource(R.xml.bookmarkpref);

		((ListPreference) findPreference(getString(R.string.pref_load)))
				.setOnPreferenceChangeListener(onPreferenceChangeListener);

		((ListPreference) findPreference(getString(R.string.pref_delete)))
				.setOnPreferenceChangeListener(onPreferenceChangeListener);

		EditTextPreference editTextPreference = (EditTextPreference) findPreference(getString(R.string.pref_add));
		editTextPreference
				.setOnPreferenceChangeListener(onPreferenceChangeListener);

		defaultSharedPreferences = PreferenceManager
				.getDefaultSharedPreferences(this);

		this.dumpPreferences(defaultSharedPreferences, "default Pref");
		int preferenceNum = defaultSharedPreferences.getInt("preferenceNum",
				9999);

		manageListPref(
				preferenceNum,
				(ListPreference) findPreference(getString(R.string.pref_delete)));
		manageListPref(preferenceNum,
				(ListPreference) findPreference(getString(R.string.pref_load)));

	}

	private void manageListPref(int preferenceNum, ListPreference listPref) {
		if (preferenceNum == 9999 || preferenceNum == 0) {
			// ブックマーク無し
			listPref.setEnabled(false);
		} else {
			Set<String> keyset = defaultSharedPreferences.getAll().keySet();
			int i = 0;
			String[] entries = new String[preferenceNum];
			for (String string : keyset) {
				if (string.startsWith(getString(R.string.prefix_prefname))) {
					entries[i] = defaultSharedPreferences.getString(string,
							"error");
					if (i != preferenceNum) {
						i++;
					} else {
						break;
					}
				}
			}

			listPref.setEntries(entries);
			listPref.setEntryValues(entries);
			listPref.setEnabled(true);
		}
	}

	OnPreferenceChangeListener onPreferenceChangeListener = new OnPreferenceChangeListener() {

		public boolean onPreferenceChange(Preference preference, Object newValue) {
			// displayToast(preference.getKey() + "changed");

			if (preference.getKey().equals(getString(R.string.pref_add))) {

				if (newValue == null || ((String) newValue).length() == 0) {
					displayToast(R.string.error_nobookmark_name);
					return false;
				}
				Set<String> entryset = defaultSharedPreferences.getAll()
						.keySet();
				for (Iterator<String> iterator = entryset.iterator(); iterator
						.hasNext();) {
					if (iterator.next().endsWith((String) newValue)) {
						displayToast(R.string.error_same_bookmarkname);
						return false;
					}

				}

				// 新しいpreferenceに値を移す
				SharedPreferences newPreference = getSharedPreferences(
						(String) newValue, MODE_PRIVATE);

				SharedPreferences.Editor newPrefEditor = newPreference.edit();

				// 書き込み
				// entryset = defaultSharedPreferences.getAll().keySet();

				for (String key : entryset) {

					// prefname_とprefenceNumはdefaultのみで管理するので書き込まない
					if (key.startsWith(getString(R.string.prefix_prefname))
							|| key.equals(getString(R.string.pref_preferenceNum))
							|| key.equals(getString(R.string.pref_orientation))) {
						continue;
					}
					try {
						// 文字列を書き込む
						newPrefEditor.putString(key,
								defaultSharedPreferences.getString(key, "999"));

					} catch (ClassCastException e) {
						try {

							// 数字を書き込む
							newPrefEditor.putInt(key,
									defaultSharedPreferences.getInt(key, 99));
						} catch (ClassCastException e2) {
							// Booleanを書き込む
							newPrefEditor.putBoolean(key,
									defaultSharedPreferences.getBoolean(key,
											false));
						}
					}
				}

				newPrefEditor.commit();

				// defaultPrefの更新
				Editor defaultPrefEditer = defaultSharedPreferences.edit();
				// 値が存在しなければ1
				int preferenceNum = defaultSharedPreferences.getInt(
						getString(R.string.pref_preferenceNum), 0) + 1;

				defaultPrefEditer.putInt(
						getString(R.string.pref_preferenceNum), preferenceNum);
				defaultPrefEditer.putString(getString(R.string.prefix_prefname)
						+ (String) newValue, (String) newValue);

				defaultPrefEditer.commit();
				manageListPref(
						defaultSharedPreferences.getInt(
								getString(R.string.pref_preferenceNum), 0),
						(ListPreference) findPreference(getString(R.string.pref_delete)));
				manageListPref(
						defaultSharedPreferences.getInt(
								getString(R.string.pref_preferenceNum), 0),
						(ListPreference) findPreference(getString(R.string.pref_load)));
				dumpPreferences(defaultSharedPreferences, "now pref");
				dumpPreferences(newPreference, "saved pref");

				// newPreference = getSharedPreferences(
				// ((EditTextPreference) preference).getText(),
				// MODE_PRIVATE);
				// newPrefEditor.clear();
				// newPrefEditor.commit();
				// defaultPrefEditer.remove("preferenceName0");
				// defaultPrefEditer.remove("preferenceName1");
				// defaultPrefEditer.remove("preferenceName2");
				// defaultPrefEditer.remove("preferenceName3");
				// defaultPrefEditer.remove("preferenceName4");
				// defaultPrefEditer.remove("preferenceNum");
				// defaultPrefEditer.remove("prefname_" + (String) newValue);
				//
				// defaultPrefEditer.commit();
				displayToast(newValue
						+ getString(R.string.message_bookmark_added));
			} else if (preference.getKey().equals(
					getString(R.string.pref_delete))) {
				// Preferenceを取得
				SharedPreferences newPreference = getSharedPreferences(
						(String) newValue, MODE_APPEND);

				dumpPreferences(newPreference, "delete pref");

				// 削除
				newPreference.edit().clear().commit();

				Editor defaultEditor = defaultSharedPreferences.edit();
				defaultEditor.remove(getString(R.string.prefix_prefname)
						+ (String) newValue);
				defaultEditor.putInt(getString(R.string.pref_preferenceNum),
						defaultSharedPreferences.getInt(
								getString(R.string.pref_preferenceNum), 1) - 1);
				defaultEditor.commit();

				manageListPref(defaultSharedPreferences.getInt(
						getString(R.string.pref_preferenceNum), 0),
						(ListPreference) preference);
				manageListPref(
						defaultSharedPreferences.getInt(
								getString(R.string.pref_preferenceNum), 0),
						(ListPreference) findPreference(getString(R.string.pref_load)));

				// ListPreference deletePref = (ListPreference) preference;
				//
				// Set<String> keyset =
				// defaultSharedPreferences.getAll().keySet();
				// int i = 0;
				// int preferenceNum = defaultSharedPreferences.getInt(
				// "preferenceNum", 0);
				//
				// if (((ListPreference) preference).getEntryValues().length ==
				// 0) {
				// preference.setEnabled(false);
				//
				// }
				dumpPreferences(defaultSharedPreferences, "now pref");
				displayToast(newValue
						+ getString(R.string.message_bookmark_deleteed));
				// else {
				//
				// String[] entries = new String[preferenceNum];
				// for (String string : keyset) {
				// if (string.startsWith("prefname_")) {
				// entries[i] = defaultSharedPreferences.getString(
				// string, "error");
				// if (i != preferenceNum) {
				// i++;
				// } else {
				// break;
				// }
				// }
				// }
				//
				// deletePref.setEntries(entries);
				// deletePref.setEntryValues(entries);
				// deletePref.setEnabled(true);
				//
				// }

			} else if (preference.getKey()
					.equals(getString(R.string.pref_load))) {
				SharedPreferences newPreference = getSharedPreferences(
						(String) newValue, MODE_APPEND);

				dumpPreferences(newPreference, "newPreference");

				Set<String> entryset = newPreference.getAll().keySet();

				SharedPreferences.Editor defaultPrefEditor = defaultSharedPreferences
						.edit();

				for (String key : entryset) {

					try {
						// 文字列を書き込む
						defaultPrefEditor.putString(key,
								newPreference.getString(key, "999"));

					} catch (ClassCastException e) {
						try {

							// 数字を書き込む
							defaultPrefEditor.putInt(key,
									newPreference.getInt(key, 99));
						} catch (Exception e2) {
							// Booleanを書き込む
							defaultPrefEditor.putBoolean(key,
									defaultSharedPreferences.getBoolean(key,
											false));
						}
					}
				}
				defaultPrefEditor.commit();

				dumpPreferences(defaultSharedPreferences,
						"defaultSharedPreferences");
				displayToast(newValue
						+ getString(R.string.message_bookmark_loaded));

			}
			return false;
		}
	};

	private void displayToast(String msg) {
		Toast.makeText(this, msg, Toast.LENGTH_SHORT).show();

	}

	private void displayToast(int resId) {
		displayToast(getString(resId));

	}

	private void dumpPreferences(SharedPreferences preferences, String message) {
		System.out.println(message);
		Set<String> entryset = preferences.getAll().keySet();

		for (String key : entryset) {
			try {
				System.out.println("key:" + key + " value:"
						+ preferences.getString(key, "999"));

			} catch (ClassCastException e) {
				try {
					System.out.println("key:" + key + " value:"
							+ preferences.getInt(key, 999));

				} catch (Exception e2) {
					System.out.println("key:" + key + " value:"
							+ preferences.getBoolean(key, false));
				}

			}
		}

	}

}
