/**
 *  $Id: HLProtocol.java,v 1.20 2001/10/08 22:03:28 groomed Exp $ Copyright (C)
 *  1998-2001 groomed <groomed@users.sourceforge.net> This program is free
 *  software; you can redistribute it and/or modify it under the terms of the
 *  GNU General Public License as published by the Free Software Foundation;
 *  either version 2 of the License, or (at your option) any later version. This
 *  program is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 *  details. You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package pharerouge.hotline;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.EOFException;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Hashtable;
import java.util.StringTokenizer;

import pharerouge.utils.DebuggerOutput;
import pharerouge.utils.EncodingUtil;
import pharerouge.utils.Sortable;
import pharerouge.utils.TextUtils;
import pharerouge.utils.ToArrayConverters;

/**
 *  This class specifies Hotline protocol constants and encapsulates Hotline
 *  protocol packets using inner classes.<p>
 *
 *  This file started out as a straightforward rip from hx. And although it's
 *  evolved since then, props where props are due.
 */
// 2002.12.25 MackyBee took NOTE: 
// Instantiatable HLProtocol may be not needed. 
public class HLProtocol {

    /**  The Hotline dir separator; usually : */
    public final static char DIR_SEPARATOR = ':';
    /**  The default TCP port for a Hotline server. */
    public final static short HTLS_TCPPORT = 5500;

    /**  The default TCP port for a Hotline tracker. */
    public final static short HTRK_TCPPORT = 5498;
    /**  Description of the Field */
    public final static short HTRK_UDPPORT = 5499;
    
    // 2003.01.08 MackyBee deactivated. 
    /**  Description of the Field */
    //public final static int SOCK_RECEIVE_BUF_SIZE = 16 * 1024;

    // 2003.01.08 MackyBee deactivated. 
    /**  Description of the Field */
    //public final static int SOCK_SEND_BUF_SIZE = 16 * 1024;

    /**
     *  Whether or not to dump hexdumps from the packet payloads to standard
     *  output as they are read or written.
     */
    public final static boolean hexdumpPayload = false;

    /* 2002.06.05 MackyBee added. */
    /**  Description of the Field */
    public final static int TYPE_BUNDLE = 2;
    /**  Description of the Field */
    public final static int TYPE_CATEGORY = 3;
    /* 2002.06.17 MackyBee added. */
    /**  Description of the Field */
    public final static short DELETE_ARTICLE_WITH_REPLIES = 0x0001;
    /**  Description of the Field */
    public final static short DELETE_ARTICLE_ONLY_ONE = 0x0000;

    /**
     *  If this is false, the file send routines will wait for the receiving end
     *  to acknowledge reception of the last block. This is the default, and the
     *  way Hotline does it. It's the safest, and pretty much guarantees that a
     *  transfer was actually completed. It's also very slow, especially when
     *  transferring many little files. So PhareRouge supports asynchronous file
     *  transfers. This way, the transfer routines just assume that everything
     *  is OK when the last block has been written. This is slightly less safe,
     *  but much faster. The problem with this is that it's incompatible with
     *  Hotline software, and you'll get "Transfer timeout!" errors from Hotline
     *  servers that you upload to. Or, if you are running a PhareRouge server
     *  with this set to true, your users will get them when they download.
     */
    public static boolean asynchronousFileTransfers = false;

    
    /**  Description of the Field */
    // 2002.12.23 MackyBee modified. Removed "static".
    // 2002.12.02 MackyBee added.
    protected String encoding = "ISO-8859-1";

    final static short DL_DATA_FORK = 2;
    final static short DL_WHOLE_FILE = 0;

    final static short HTLC_DATA_BAN = 0x0071;
    final static short HTLC_DATA_CHAT = 0x0065;
    final static short HTLC_DATA_CHAT_REF = 0x0072;
    final static short HTLC_DATA_CHAT_SUBJECT = 0x0073;

    // 2002.10.22 MackyBee PROVISIONALLY, EXPERIMENTALLY added.
    // Note: This value is same as HTLC_DATA_BAN.
    final static short HTLC_DATA_COMM_RESTRICTION = 0x0071;

    /* Added 2002.02.25 by MackyBee */
    final static short HTLC_DATA_CLIENT_VERSION = 0x00a0;
    final static short HTLC_DATA_DIR = 0x00ca;
    final static short HTLC_DATA_DIR_RENAME = 0x00d4;
    final static short HTLC_DATA_FILE = 0x00c9;
    final static short HTLC_DATA_FILENAME = 0x0000007b;
    final static short HTLC_DATA_FILE_RENAME = 0x00d3;

    final static short HTLC_DATA_ICON = 0x0068;
    final static short HTLC_DATA_LISTDIR = 0x00ca;
    final static short HTLC_DATA_LOGIN = 0x0069;
    final static short HTLC_DATA_MSG = 0x0065;
    final static short HTLC_DATA_NEWS_POST = 0x0065;
    final static short HTLC_DATA_NICK = 0x0066;
    final static short HTLC_DATA_OPTION = 0x006d;
    final static short HTLC_DATA_PASSWORD = 0x006a;
    final static short HTLC_DATA_RESUME = 0x000000cc;
    final static short HTLC_DATA_RFLT = 0x00cb;
    final static short HTLC_DATA_SOCKET = 0x0067;
    final static short HTLC_DATA_XFERSIZE = 0x006c;
    final static int HTLC_HDR_ACCOUNT_CREATE = 0x0000015e;
    final static int HTLC_HDR_ACCOUNT_DELETE = 0x0000015f;
    final static int HTLC_HDR_ACCOUNT_MODIFY = 0x00000161;
    final static int HTLC_HDR_ACCOUNT_READ = 0x00000160;
    final static int HTLC_HDR_AGREE = 0x0079;
    final static int HTLC_HDR_CHAT = 0x00000069;

    /* 2002.06.02 modified int -> short. 2002.06.01 MackyBee added */
    final static short HTLC_DATA_NEWS_COMPONENT = 0x0145;
    /* 2002.06.04 MackyBee added */
    final static short HTLC_DATA_NEWS_ARTICLE_INDEX = 0x0146;
    /* 2002.06.04 MackyBee added */
    final static short HTLC_DATA_NEWS_ARTICLE_MIME_TYPE = 0x0147;
    /* 2002.06.15 MackyBee added */
    final static short HTLC_DATA_NEWS_BUNDLE_TITLE = 0x00c9;
    /* 2002.06.15 MackyBee added */
    final static short HTLC_DATA_NEWS_CATEGORY_TITLE = 0x0142;
    /* 2002.06.15 MackyBee added */
    final static short HTLC_DATA_NEWS_ARTICLE_TITLE = 0x0148;
    /* 2002.06.15 MackyBee added */
    final static short HTLC_DATA_NEWS_UNKNOWN_01 = 0x014e;
    /* 2002.06.15 MackyBee added */
    final static short HTLC_DATA_NEWS_DELETE_TYPE = 0x0151;

    /*  Added 2002.03.14 MackyBee
		In case connecting to remote servers, the HL client sometimes send
		this id packet periodically.
		And, on some user environments, that function avoid unexpected
		disconnection.
		The variable name is tentative. */
    final static int HTLC_HDR_CONNECTION_ALIVE = 0x01f4;

    final static int HTLC_HDR_DIR_CREATE = 0x000000cd;
    final static int HTLC_HDR_DIR_LIST = 0x000000c8;
    final static int HTLC_HDR_FILE_DELETE = 0x000000cc;
    final static int HTLC_HDR_FILE_GET = 0x000000ca;
    final static int HTLC_HDR_FILE_GETINFO = 0x000000ce;
    final static int HTLC_HDR_FILE_MAKE_ALIAS = 0x000000d1;
    final static int HTLC_HDR_FILE_MOVE = 0x000000d0;
    final static int HTLC_HDR_FILE_PUT = 0x000000cb;
    final static int HTLC_HDR_FILE_SETINFO = 0x000000cf;
    final static int HTLC_HDR_LOGIN = 0x0000006b;
    final static int HTLC_HDR_MSG = 0x0000006c;
    final static int HTLC_HDR_NEWS_GET = 0x00000065;
    final static int HTLC_HDR_NEWS_POST = 0x00000067;
    final static int HTLC_HDR_PRIVCHAT_CREATE = 0x00000070;
    final static int HTLC_HDR_PRIVCHAT_DECLINE = 0x00000072;
    final static int HTLC_HDR_PRIVCHAT_INVITE = 0x00000071;
    final static int HTLC_HDR_PRIVCHAT_JOIN = 0x00000073;
    final static int HTLC_HDR_PRIVCHAT_LEAVE = 0x00000074;
    final static int HTLC_HDR_PRIVCHAT_SUBJECT = 0x00000078;
    /*
		Added experimentally 2002.02.25 by MackyBee
		The variable is same as HTLC_HDR_USER_CHANGE
	*/
    final static int HTLC_HDR_SEND_OLD_USERINFO = 0x0130;
    /*
		Added experimentally 2002.02.24 by MackyBee
		The variable is same as HTLC_HDR_AGREE.
	*/
    final static int HTLC_HDR_SEND_USERINFO = 0x0079;

    /* Should sort this out properly some time. */
    final static int HTLC_HDR_USER_CHANGE = 0x00000130;
    final static int HTLC_HDR_USER_GETINFO = 0x0000012f;
    final static int HTLC_HDR_USER_KICK = 0x0000006e;
    final static int HTLC_HDR_USER_LIST = 0x0000012c;

    /* 2002.05.27 MackyBee added */
    final static int HTLC_HDR_GET_NEWS_BUNDLE = 0x00000172;
    /* 2002.06.04 MackyBee added */
    final static int HTLC_HDR_GET_NEWS_THREADS = 0x00000173;
    /* 2002.06.04 MackyBee added */
    final static int HTLC_HDR_GET_NEWS_ARTICLE = 0x00000190;
    /* 2002.06.15 MackyBee added */
    final static int HTLC_HDR_CREATE_NEWS_BUNDLE = 0x0000017d;
    /* 2002.06.15 MackyBee added */
    final static int HTLC_HDR_CREATE_NEWS_CATEGORY = 0x0000017e;
    /* 2002.06.15 MackyBee added */
    final static int HTLC_HDR_POST_NEWS_ARTICLE = 0x0000019a;
    /* 2002.06.17 MackyBee added */
    final static int HTLC_HDR_DELETE_NEWS_ARTICLE = 0x0000019b;
    final static int HTLC_HDR_DELETE_NEWS_CONTAINER = 0x0000017c;

    /*
		Note 2002.02.25 by MackyBee
		The following String variables are not encorded to byte correctly
		on some encoding platforms, so it uses byte[] instead of them.
	 */
    /* Modified 2002.02.14 by MackyBee */
    //static final String HTLC_MAGIC = "TRTPHOTL\0\1\0\2";
    final static byte[] HTLC_MAGIC =
        {0x54, 0x52, 0x54, 0x50, 0x48, 0x4f, 0x54, 0x4c,
        0x00, 0x01, 0x00, 0x02};
    final static short HTLC_MAGIC_LEN = 12;
    final static short HTLS_DATA_AGREEMENT = 0x0065;
    final static short HTLS_DATA_AUTO_AGREE = 0x009a;

    final static short HTLS_DATA_CHAT = 0x0065;
    final static short HTLS_DATA_CHAT_REF = 0x0072;
    final static short HTLS_DATA_CHAT_SUBJECT = 0x0073;
    final static short HTLS_DATA_COLOUR = 0x0070;
    final static short HTLS_DATA_FILE_COMMENT = 0x00d2;
    final static short HTLS_DATA_FILE_CREATOR = 0x00ce;
    final static short HTLS_DATA_FILE_DATE_CREATED = 0x00d0;
    final static short HTLS_DATA_FILE_DATE_MODIFIED = 0x00d1;
    final static short HTLS_DATA_FILE_ICON = 0x00d5;
    final static short HTLS_DATA_FILE_LIST = 0x00c8;
    final static short HTLS_DATA_FILE_NAME = 0x00c9;
    final static short HTLS_DATA_FILE_SIZE = 0x00cf;
    final static short HTLS_DATA_FILE_TYPE = 0x00cd;
    final static short HTLS_DATA_HTXF_QUEUEPOS = 0x0074;
    final static short HTLS_DATA_HTXF_REF = 0x006b;
    final static short HTLS_DATA_HTXF_RFLT = 0x00cb;
    final static short HTLS_DATA_HTXF_SIZE = 0x006c;
    final static short HTLS_DATA_ICON = 0x0068;
    final static short HTLS_DATA_LOGIN = 0x0069;
    final static short HTLS_DATA_MSG = 0x0065;
    final static short HTLS_DATA_NEWS = 0x0065;
    final static short HTLS_DATA_NICK = 0x0066;
    final static short HTLS_DATA_OPTION = 0x006d;
    final static short HTLS_DATA_PASSWORD = 0x006a;
    final static short HTLS_DATA_PRIVILEGES = 0x006e;

    /* 2002.05.27 MackyBee added */
    final static short HTLS_DATA_BUNDLE_LIST = 0x0143;
    /* 2002.06.03 MackyBee added */
    final static short HTLS_DATA_NEWS_THREAD = 0x0141;
    /* 2002.06.05 MackyBee added */
    final static short HTLS_DATA_NEWS_ARTICLE = 0x014d;

    /* Added 2002.02.28 MackyBee */
    final static short HTLS_DATA_SERVER_NAME = 0x00a2;

    final static short HTLS_DATA_SOCKET = 0x0067;
    final static short HTLS_DATA_TASKERROR = 0x0064;
    final static short HTLS_DATA_USER_INFO = 0x0065;
    final static short HTLS_DATA_USER_LIST = 0x012c;

    final static short HTLS_DATA_VERSION = 0x00a0;

    final static int HTLS_HDR_ACCOUNT_INFO = 0x0162;
    final static int HTLS_HDR_AGREEMENT = 0x0000006d;
    final static int HTLS_HDR_CHAT = 0x0000006a;
    final static int HTLS_HDR_CHAT_INVITE = 0x00000071;
    final static int HTLS_HDR_CHAT_SUBJECT = 0x00000077;
    final static int HTLS_HDR_CHAT_USER_CHANGE = 0x00000075;
    final static int HTLS_HDR_CHAT_USER_LEAVE = 0x00000076;
    final static int HTLS_HDR_MSG = 0x00000068;
    final static int HTLS_HDR_NEWS_POST = 0x00000066;
    final static int HTLS_HDR_POLITEQUIT = 0x0000006f;
    final static int HTLS_HDR_QUEUEPOS = 0x00d3;

    final static int HTLS_HDR_TASK = 0x00010000;
    final static int HTLS_HDR_USERDISCONNECT = 0x0000012e;
    final static int HTLS_HDR_USER_CHANGE = 0x0000012d;

    final static int HTLS_HDR_USER_LEAVE = 0x0000012e;
    /* Modified 2002.02.14 by MackyBee */
    //static final String HTLS_MAGIC = "TRTP\0\0\0\0";
    final static byte[] HTLS_MAGIC =
        {0x54, 0x52, 0x54, 0x50, 0x00, 0x00, 0x00, 0x00};
    final static short HTLS_MAGIC_LEN = 8;

    /* HOPE (Hotline Open Protocol Extensions) */
    final static short HTLX_DATA_APP_ID = 0xe01;
    final static short HTLX_DATA_APP_STRING = 0xe02;
    final static short HTLX_DATA_MAC_ALGORITHM = 0xe04;
    final static short HTLX_DATA_SESSION_KEY = 0xe03;
    /* Modified 2002.02.14 by MackyBee */
    //static final String HTRK_MAGIC = "HTRK\0\1";
    final static byte[] HTRK_MAGIC =
        {0x48, 0x54, 0x52, 0x4b, 0x00, 0x01};

    final static short HTRK_MAGIC_LEN = 6;
    final static String HTXF_MAGIC = "HTXF";
    final static int HTXF_MAGIC_INT = 0x48545846;
    final static short HTXF_MAGIC_LEN = 4;
    final static short HTXF_TCPPORT = 5501;

    final static short SIZEOF_DATACOMPONENT = (4);

    final static short SIZEOF_PACKETHEADER = (22);

    /* Added Experimentally 2002.03.08 MackyBee */
    PeriodicalPacketSender pps;

    private SimpleDateFormat dateFormatter;

    /**  Constructor for the HLProtocol object */
    public HLProtocol() {
        dateFormatter = new SimpleDateFormat("EEE, MMM dd, yyyy, kk:mm:ss");
    }

    /**
     * @param  encoding  Specifies character encoding used by String.getBytes(
     *      String) method. This setting is also used by ComponentFactory class.
     */
    // 2002.12.02 MackyBee added.
    public HLProtocol(String encoding) {
        this();
        this.encoding = encoding;
        // 2002.12.21 MackyBee added. 
        //ComponentFactory.setEncoding(encoding);
    }

    // 2002.12.23 MackyBee modified. Removed "static".
    // 2002.12.21 MackyBee added. 
    public void setEncoding(String encoding) {
        if (encoding != null) {
            this.encoding = encoding;
        }
    }
    
    // 2002.12.23 MackyBee modified. Removed "static".
    // 2002.12.22 MackyBee added. 
    public String getEncoding() {
        return (encoding);
    }
    
    /**
     *  Translates a Calendar into the corresponding Hotline date stamp.
     *
     * @param  calendar  the date to translate.
     * @return           8 byte Hotline date stamp.
     */
    public static byte[] calendarToHotlineDate(Calendar calendar) {

        byte[] hotlineDate = new byte[8];

        /* The 1.2.3 (Macintosh) version of the Hotline client
           seems broken in how it handles dates, but strangely the
           server is not. In any case this will work for Tube (and
           Tube also works correctly with the 1.2.3 server). */
        byte[] epochBytes = ToArrayConverters.shortToByteArray(1970);
        byte[] secondsBytes = ToArrayConverters.intToByteArray(
            (int) (calendar.getTime().getTime() / 1000));

        hotlineDate = new byte[8];
        hotlineDate[0] = epochBytes[0];
        hotlineDate[1] = epochBytes[1];
        hotlineDate[2] = 0;
        hotlineDate[3] = 0;
        hotlineDate[4] = secondsBytes[0];
        hotlineDate[5] = secondsBytes[1];
        hotlineDate[6] = secondsBytes[2];
        hotlineDate[7] = secondsBytes[3];

        return hotlineDate;
    }

    /**
     *  From an 8 byte array signifying a Hotline date stamp, calculates the
     *  date and returns it.<P>
     *
     *  The format of a Hotline date stamp is:<P>
     *
     *  <PRE>
     * bytes    description
     * 0 - 1    epoch (e.g. 1904)
     * 2 - 3    unknown (zero)
     * 4 - 7    seconds since epoch</PRE>
     *
     * @param  hotlineDate                an 8 byte Hotline date stamp.
     * @return                            a Calendar date.
     * @throws  IllegalArgumentException  if hotlineDate is not an 8 byte array.
     */
    public static Calendar hotlineDateToCalendar(byte[] hotlineDate) {

        if (hotlineDate.length != 8) {
            throw new IllegalArgumentException("Hotline date is not 8 bytes");
        }

        Calendar calendar = Calendar.getInstance();
        byte[] epoch = new byte[2];
        epoch[0] = hotlineDate[0];
        epoch[1] = hotlineDate[1];

        /* Don't know what hotlineDate[2] and hotlineDate[3] represent. */
        byte[] seconds = new byte[4];
        seconds[0] = hotlineDate[4];
        seconds[1] = hotlineDate[5];
        seconds[2] = hotlineDate[6];
        seconds[3] = hotlineDate[7];
        int nEpoch = ToArrayConverters.byteArrayToInt(epoch);
        long nSeconds = ToArrayConverters.byteArrayToLong(seconds);
        int days = (int) (nSeconds / 86400);
        /* 86400 = 24 * 3600 */
        calendar.setLenient(true);
        calendar.clear();
        calendar.set(Calendar.YEAR, nEpoch);
        calendar.add(Calendar.DAY_OF_MONTH, days);
        calendar.add(Calendar.SECOND, (int) nSeconds - (days * 86400));

        return calendar;
    }

    /**
     *  Description of the Method
     *
     * @param  b  Description of the Parameter
     * @return    Description of the Return Value
     */
    static byte[] invert(byte[] b) {
        byte[] encoded = b;

        for (int i = 0; i < b.length; i++) {
            encoded[i] = (byte) (255 - encoded[i]);
        }

        return encoded;
    }

    /**
     *  Constructs a DataComponent from the given DataInputStream. For any given
     *  DataComponent on the input, this method creates and returns the most
     *  specific subclass of DataComponent that it knows about.
     *
     * @param  din  the input stream to construct from.
     * @return      most specific subclass of DataComponent that is possible.
     */
    // 2002.12.22 MackyBee modified. Moved this from other class. 
    HLProtocol.DataComponent createComponent(DataInputStream din) {

        short type;

        try {
            /* Read the type of this component, then pass object
               creation on to the appropriate constructors, or create
               a generic DataComponent if there is no specialized
               object to represent the given type. */
            type = din.readShort();

            switch (type) {

                case HLProtocol.HTLS_DATA_HTXF_RFLT:
                    return new ResumeTransferComponent(din);
                case HLProtocol.HTLS_DATA_FILE_LIST:
                    return new FileListComponent(din);
                case HLProtocol.HTLS_DATA_USER_LIST:
                    return new UserListComponent(din);
                case HLProtocol.HTLC_DATA_DIR:
                case HLProtocol.HTLC_DATA_DIR_RENAME:
                    return new PathComponent(type, din);
                case HLProtocol.HTLC_DATA_FILE:
                    return new FileComponent(din);
                case HLProtocol.HTLS_DATA_FILE_DATE_CREATED:
                case HLProtocol.HTLS_DATA_FILE_DATE_MODIFIED:
                    return new DateComponent(type, din);
                /* 2002.06.01 MackyBee note:
                   There is no ID difference bundle and category.
                   2002.05.31 MackyBee added */
                case HLProtocol.HTLS_DATA_BUNDLE_LIST:
                    return new NewsBundleListComponent(din);
                /* 2002.06.04 MackyBee added */
                case HLProtocol.HTLS_DATA_NEWS_THREAD:
                    return new NewsThreadListComponent(din);
                /* 2002.06.05 MackyBee added */
                case HLProtocol.HTLC_DATA_NEWS_ARTICLE_MIME_TYPE:
                    return new NewsArticleMIMETypeComponent(din);
                /* 2002.06.07 MackyBee added */
                case HLProtocol.HTLS_DATA_NEWS_ARTICLE:
                    return new NewsArticleComponent(din);
                case HLProtocol.HTLC_DATA_NEWS_ARTICLE_INDEX:
                    return new NewsArticleIndexComponent(din);
                default:
                    return new DataComponent(type, din);
            }

        } catch (EOFException e) {
            /* This is to compensate for a bug in Tube 1.0b2. Tube
               1.0b2 doesn't get the PacketHeader.hc field right. See
               also the PacketHeader constructor. */
            return null;
        } catch (IOException e) {
            e.printStackTrace();
        }

        return null;
    }

    /**
     *  Returns an array containing one or two DataComponents which describing
     *  the requested path. The number of DataComponents returns depends on the
     *  value of isFile: if it is true, 2 components are returned.
     *
     * @param  path             the Hotline path.
     * @param  dataType         {@link HLProtocol.HTLC_DATA_DIR} or {@link
     *      HLProtocol.HTLC_DATA_DIR_RENAME}.
     * @param  isFile           true if the path refers to a file.
     * @return                  Description of the Return Value
     * @exception  IOException  Description of the Exception
     */
    // 2002.12.22 MackyBee modified. Moved this from other class. 
    HLProtocol.DataComponent[] createPathComponents(String path,
        short dataType,
        boolean isFile) throws IOException {

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[isFile ? 2 : 1];

        if (isFile) {
            String file = path.substring(path.lastIndexOf(
                HLProtocol.DIR_SEPARATOR) + 1);
            dataComponents[1] = new FileComponent(file);

            /* 2002.04.04 MackyBee modified -> return to original */
            //path = path.substring(0, path.getBytes(encoding).length -
            //    file.getBytes(encoding).length);
            path = path.substring(0, path.length() - file.length());
        }

        dataComponents[0] = new PathComponent(dataType, path);

        return dataComponents;
    }
    
    /**  Describes a Hotline user account. */
    public class AccountInfo {

        /**  Description of the Field */
        public final static long CANNOT_BE_DISCONNECTED = 0x00010000L;
        /**  Description of the Field */
        public final static long CAN_COMMENT_FILES = 0x08000000L;
        /**  Description of the Field */
        public final static long CAN_COMMENT_FOLDERS = 0x04000000L;
        /**  Description of the Field */
        public final static long CAN_CREATE_ACCOUNTS = 0x00000200L;
        /**  Description of the Field */
        public final static long CAN_CREATE_FOLDERS = 0x00000004L;
        /**  Description of the Field */
        public final static long CAN_DELETE_ACCOUNTS = 0x00000100L;
        /**  Description of the Field */
        public final static long CAN_DELETE_FILES = 0x00000080L;
        /**  Description of the Field */
        public final static long CAN_DELETE_FOLDERS = 0x00000002L;
        /**  Description of the Field */
        public final static long CAN_DISCONNECT_USERS = 0x00020000L;
        /**  Description of the Field */
        public final static long CAN_DOWNLOAD_FILES = 0x00000020L;
        /**  Description of the Field */
        public final static long CAN_GET_USER_INFO = 0x80000000L;
        /**  Description of the Field */
        public final static long CAN_MAKE_ALIASES = 0x01000000L;
        /**  Description of the Field */
        public final static long CAN_MODIFY_ACCOUNTS = 0x00400000L;
        /**  Description of the Field */
        public final static long CAN_MOVE_FILES = 0x00000008L;
        /**  Description of the Field */
        public final static long CAN_MOVE_FOLDERS = 0x00008000L;
        /**  Description of the Field */
        public final static long CAN_POST_NEWS = 0x00040000L;
        /**  Description of the Field */
        public final static long CAN_READ_ACCOUNTS = 0x00800000L;
        /**  Description of the Field */
        public final static long CAN_READ_CHAT = 0x00004000L;
        /**  Description of the Field */
        public final static long CAN_READ_NEWS = 0x00080000L;
        /**  Description of the Field */
        public final static long CAN_RENAME_FILES = 0x00000010L;

        /**  Constants for masking the {@link #privileges} field. */
        public final static long CAN_RENAME_FOLDERS = 0x00000001L;
        /**  Description of the Field */
        public final static long CAN_SEND_CHAT = 0x00002000L;
        /**  Description of the Field */
        public final static long CAN_UPLOAD_ANYWHERE = 0x40000000L;
        /**  Description of the Field */
        public final static long CAN_UPLOAD_FILES = 0x00000040L;
        /**  Description of the Field */
        public final static long CAN_USE_ANY_NAME = 0x20000000L;
        /**  Description of the Field */
        public final static long CAN_VIEW_DROPBOXES = 0x02000000L;
        /**  Description of the Field */
        public final static long DONT_SHOW_AGREEMENT = 0x10000000L;

        /**
         *  The home directory for this account (not strictly a part of the
         *  Hotline protocol, and may be null).
         */
        public String homeDirectory;

        /**  The login for this account. */
        public String login;

        /**  The nickname for this account. */
        public String nick;

        /**
         *  The password for this account. Included for completeness only, the
         *  Hotline server never actually sends the password.
         */
        public String password;

        /**
         *  The privileges for this account: <pre>
         * bit 1: can rename folders
         * bit 2: can delete folders
         * bit 3: can create folders
         * bit 4: can move files
         * bit 5: can rename files
         * bit 6: can download files
         * bit 7: can upload files
         * bit 8: can delete files
         * bit 9: can delete users
         * bit 10: can create users
         * bit 11: reserved
         * bit 12: reserved
         * bit 13: reserved
         * bit 14: can send chat
         * bit 15: can read chat
         * bit 16: can move folders
         * bit 17: cannot be disconnected
         * bit 18: can disconnect users
         * bit 19: can post news
         * bit 20: can read news
         * bit 21: reserved
         * bit 22: reserved
         * bit 23: can modify users
         * bit 24: can read users
         * bit 25: can make aliases
         * bit 26: can view drop boxes
         * bit 27: can comment folders
         * bit 28: can comment files
         * bit 29: don't show agreement
         * bit 30: can use any name
         * bit 31: can upload anywhere
         * bit 32: can get user info
         * </pre> {@see pharerouge.hotline.HLProtocol.AccountInfo#CAN_RENAME_FOLDERS}
         */
        public long privileges;

        /**
         *  Creates the object with the specified values.
         *
         * @param  l   login.
         * @param  n   nickname.
         * @param  p   password.
         * @param  pr  privileges.
         * @param  h   home directory for this account.
         */
        public AccountInfo(String l, String n, String p, long pr, String h) {
            login = l;
            nick = n;
            password = p;
            privileges = pr;
            homeDirectory = h;
        }

        /**  Creates the object with empty fields. */
        public AccountInfo() {
            this("", "", "", 0, "");
        }

        /**
         *  Description of the Method
         *
         * @return    Description of the Return Value
         */
        public String toString() {
            return "HLProtocol.AccountInfo[login = " + login
                 + ", password = <not shown>, nick = " + nick
                 + ", privileges = " + privileges + ", homeDirectory = "
                 + homeDirectory + "]";
        }

    }

    /**  Describes info for a file. */
    public class FileInfo {

        /**  The comment of the file. */
        public String comment;

        /**  The creation date of the file. */
        public Calendar created;

        /**  The creator of the file. */
        public String creator;

        /**  The "icon" of the file (usually 4 character code). */
        public String icon;

        /**  The date the file was last modified. */
        public Calendar modified;

        /**  The name of the file. */
        public String name;

        /**  The size of the file. */
        public long size;

        /**  The type of the file. */
        public String type;

        /**
         *  Description of the Method
         *
         * @return    Description of the Return Value
         */
        public String toString() {
            return "FileInfo[" + name + ", " + creator + ", " + type + ", "
                 + icon + ", " + comment + ", " + size + ", " +
                dateFormatter.format(created.getTime()) + ", " +
                dateFormatter.format(modified.getTime()) + "]";
        }

    }

    /**  Describes a file as a part of a file list. */
    public class FileListComponent extends DataComponent
         implements Cloneable, Sortable {

        /**  The (MacOS) file creator of the file. */
        public String fileCreator;

        /**  The name of the file. */
        public String fileName;

        /**  The size of the file. */
        public long fileSize;

        /**  The (MacOS) file type of the file. */
        public String fileType;

        /**  Unknown. */
        public int unknown;

        /**
         *  Creates a file list component with the specified parameters.
         *
         * @param  fName     the file name.
         * @param  fType     4 byte file type info.
         * @param  fCreator  4 byte file creator info.
         * @param  fSize     the file size.
         * @param  fExtra    (unknown).
         */
        public FileListComponent(String fName,
            String fType,
            String fCreator,
            long fSize,
            int fExtra) {

            super(HTLS_DATA_FILE_LIST);

            if (fType.length() != 4) {
                throw new IllegalArgumentException(
                    "File type must be 4 bytes.");
            }

            if (fCreator.length() != 4) {
                throw new IllegalArgumentException(
                    "File creator must be 4 bytes.");
            }

            fileName = fName;
            fileType = fType;
            fileCreator = fCreator;
            fileSize = fSize;
            unknown = fExtra;

            construct();
        }

        /**
         *  Interprets the given DataComponent as a FileListComponent and
         *  initializes the FileListComponent with those values.
         *
         * @param  is               Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        public FileListComponent(DataInputStream is) throws IOException {

            super(HTLS_DATA_FILE_LIST, is);

            DataInputStream din =
                new DataInputStream(new ByteArrayInputStream(data));

            byte[] ftype = new byte[4];
            byte[] fcreator = new byte[4];
            din.read(ftype, 0, 4);
            din.read(fcreator, 0, 4);
            fileSize = din.readInt();
            unknown = din.readInt();
            byte[] fname = new byte[din.readInt()];
            din.readFully(fname, 0, fname.length);


           // 2005.11.27 Gardien modified.
            fileType = new String(ftype, encoding);
            fileCreator = new String(fcreator, encoding);
            fileName = new String(fname, encoding);
/*           // 2002.11.13 MackyBee EXPERIMENTALLY modified.
              fileType = new String(ftype, EncodingUtil.defaultEncoding);
              fileCreator = new String(fcreator, EncodingUtil.defaultEncoding);
              fileName = new String(fname, EncodingUtil.defaultEncoding);
            //fileType = new String(ftype);
            //fileCreator = new String(fcreator);
            //fileName = new String(fname); 
*/
        }

        /**
         *  Description of the Method
         *
         * @return                                 Description of the Return
         *      Value
         * @exception  CloneNotSupportedException  Description of the Exception
         */
        public Object clone() throws CloneNotSupportedException {
            return super.clone();
        }

        /**
         *  Description of the Method
         *
         * @param  other  Description of the Parameter
         * @return        Description of the Return Value
         */
        public boolean smallerThan(Sortable other) {

            if (other instanceof HLProtocol.FileListComponent) {
                HLProtocol.FileListComponent otherFile =
                    (HLProtocol.FileListComponent) other;

                /* Directories precede files. */
                if (fileType.equals("fldr") && !otherFile.fileType.equals(
                    "fldr")) {
                    return true;
                }
                if (!fileType.equals("fldr") && otherFile.fileType.equals(
                    "fldr")) {
                    return false;
                }
                return fileName.compareToIgnoreCase(otherFile.fileName) < 0;
            } else {
                return fileName.compareToIgnoreCase(other.toString()) < 0;
            }
        }

        /**
         *  Description of the Method
         *
         * @return    Description of the Return Value
         */
        public String toString() {
            return "FileListComponent[" + fileType + ", "
                 + (fileType.equals("fldr") ? "" : fileCreator + ", ")
                 + fileSize + ", " + fileName + "]";
        }

        /**  Description of the Method */
        private void construct() {
            try {
                ByteArrayOutputStream bos = new ByteArrayOutputStream();
                DataOutputStream dos = new DataOutputStream(bos);

                byte[] ftype = fileType.getBytes(encoding);
                byte[] fcreator = fileCreator.getBytes(encoding);
                dos.write(ftype, 0, 4);
                dos.write(fcreator, 0, 4);
                dos.writeInt((int) fileSize);
                dos.writeInt(unknown);
                // 2002.11.11 MackyBee modified.
                // 2002.03.27 MackyBee modified.
                byte[] fnBytes = EncodingUtil.strictGetBytes(fileName);
                //byte[] fnBytes = fileName.getBytes(encoding);
                dos.writeInt(fnBytes.length);
                dos.write(fnBytes);
                //dos.writeInt(fileName.length());
                //dos.write(fileName.getBytes(encoding));
                dos.flush();

                data = bos.toByteArray();
            } catch (IOException e) {}
        }
    }

    /**  Describes a server as gotten from a tracker. */
    public class ServerInfo {

        /**  Address of server. */
        public int address;
        
        /**  The server description. */
        public String desc;

        /**  The server name. */
        public String name;
        
        /**  Number of users at server. */
        public short nusers;
        
        /**  Port of server. */
        // 2002.12.26 MackyBee modified. 
        public int port;
        //public char port;

        /**  Unknown. */
        public short unknown;

        /**  Constructs empty. */
        public ServerInfo() { }

        /**
         *  Constructs from a DataInputStream.
         *
         * @param  din              Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        public ServerInfo(DataInputStream din) throws IOException {
            read(din);
        }

        /**
         *  Gets the address attribute of the ServerInfo object
         *
         * @return    The address value
         */
        public String getAddress() {
            char[] b = ToArrayConverters.intToCharArray(address);
            return new String((int) b[0] + "." + (int) b[1] + "." 
                + (int) b[2] + "." + (int) b[3]);
        }

        /**
         *  Description of the Method
         *
         * @return    Description of the Return Value
         */
        public String toString() {
            return "ServerInfo[name = " + name + ", description = " + desc 
                + "]";
        }

        /**
         *  Description of the Method
         *
         * @param  din              Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        void read(DataInputStream din) throws IOException {

            address = din.readInt();
            // 2002.12.26 MackyBee modified. 
            port = din.readUnsignedShort();
            //port = (char) din.readShort();
            nusers = din.readShort();
            unknown = (short) din.readShort();

            /* 2002.04.14 MackyBee modified */
            byte[] n = new byte[din.readUnsignedByte()];
            //byte[] n = new byte[(short) (SignUtils.toUnsigned(din.readByte()) & 0xff)];

            din.readFully(n, 0, n.length);

            /* Explicitly specify ISO-8859-1 encoding or we may run
               into converter errors, crashing the thread. Probably
               Sun's or MS's bug ... */
            /*
				Modified by MackyBee <macky@email.plala.or.jp>.
                In order to internationalize.
                At least, on Windows 98 SE with JDK 1.3.1 and above
                version, I have never got a problem.
                Wrote on 2002.02.02
             */
            //name = new String(n, "ISO-8859-1");

            /* 2002.05.02 modifies, Modified 2002.02.02 by MackyBee */
            //name = new String(new String(n).getBytes(
            //    pharerouge.client.Main.DEFAULT_ENCODING));
            // 2002.12.26 MackyBee modified. 
            name = new String(n, encoding);
            //name = new String(n);

            /* 2002.04.14 MackyBee modified */
            byte[] d = new byte[din.readUnsignedByte()];
            //byte[] d = new byte[(short) (SignUtils.toUnsigned(din.readByte())
            //     & 0xff)];

            din.readFully(d, 0, d.length);

            //desc = new String(d, "ISO-8859-1");
            /* Modified on 2002.02.02 by MackyBee */
            // 2002.12.26 MackyBee modified. 
            desc = new String(d, encoding);
            //desc = new String(d);
        }
    }

    /**  Describes a user as a part of a user list. */
    public class UserListComponent extends DataComponent {

        // 2002.11.21 MackyBee modified. This works.
        // 2002.04.09 MackyBee added. (short) 0x12 = (int) 18
        /**  A data ID represents the head admin. */
        public final static short HEAD_ADMIN = 0x10;
        //public final static short HEAD_ADMIN = 0x12;

        // 2002.11.19 MackyBee added.
        /**  Description of the Field */
        public final static short REFUSE_PRIVATE_CHAT = 0x08;

        // 2002.11.19 MackyBee added. This is for the HLC 1.5 above function.
        /**  Description of the Field */
        public final static short REFUSE_PRIVATE_MESSAGE = 0x04;

        /**  Description of the Field */
        public final static short CAN_DISCONNECT_USERS = 0x02;

        /**  Constants to mask the {@link #clr} field. */
        public final static short HAS_BEEN_IDLE = 0x01;

        /**  Client request of refuse private message. */
        // 2002.11.20 MackyBee added.
        public final static short CR_REFUSE_P_MESSAGE = 0x01;
        /**  Client request of refuse private chat. */
        // 2002.11.20 MackyBee added.
        public final static short CR_REFUSE_P_CHAT = 0x02;

        /**  Client request of refuse private message and chat. */
        // 2002.11.21 deactivated.
        // 2002.11.20 MackyBee EXPERIMENTALLY added.
        //public final static short CR_REFUSE_P_MESSAGE_AND_CHAT = 0x03;

        /**  The color (status) of the user. */
        public short clr;

        /**  The icon number of the user. */
        public int icon;

        /**  The nick of the user. */
        public String nick = null;

        /**  The sock of the user. */
        public int sock;

        /**
         *  Constructs a UserListComponent with the given values.
         *
         * @param  s  the sock of the user.
         * @param  i  the icon number of the user.
         * @param  c  the color of the user.
         * @param  n  the nickname of the user.
         */
        UserListComponent(int s, int i, short c, String n) {
            super(HTLS_DATA_USER_LIST);

            sock = s;
            icon = i;
            clr = c;
            nick = n;

            construct();
        }

        /**
         *  Constructs a UserListComponent from an input stream from which the
         *  type short has already been read.
         *
         * @param  is               Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        UserListComponent(DataInputStream is) throws IOException {
            super(HTLS_DATA_USER_LIST, is);

            DataInputStream din =
                new DataInputStream(new ByteArrayInputStream(data));
            // 2002.04.14 MackyBee modified.
            // The previous my codes caused problems.
            sock = din.readUnsignedShort();
            icon = din.readUnsignedShort();
            // 2002.03.22 MackyBee modified experimentally
            //sock = (int) din.readShort();
            //icon = (int) din.readShort();
            //sock = (int) din.readChar();
            //icon = (int) din.readChar();

            clr = din.readShort();

            // 2002.04.14 MackyBee modified. 
            byte[] n = new byte[din.readUnsignedShort()];
            /* 2002.03.24 MackyBee noted. Here are no problems maybe. */
            /* 2002.03.22 MackyBee modified */
            /* Note: readChar depends on the Unicode. */
            //byte[] n = new byte[(int) din.readShort()];
            //byte[] n = new byte[(int) din.readChar()];

            din.readFully(n, 0, n.length);
            // 2002.12.21 MackyBee modified. 
            nick = new String(n, encoding);
            //nick = new String(n);
        }

        /**
         *  Returns a string representation of this object (the user's nick
         *  name).
         *
         * @return    nick name.
         */
        public String toString() {
            String s = "UserListComponent[icon: " + icon + ", sock: " + sock
                 + ", name: " + new String(nick);
            String state = null;

            /* 2002.04.09 MackyBee added */
            if ((clr & HEAD_ADMIN) == HEAD_ADMIN) {
                state = "HEAD_ADMIN";
            }

            if ((clr & CAN_DISCONNECT_USERS) == CAN_DISCONNECT_USERS) {
                state = "CAN_DISCONNECT_USERS";
            }

            if ((clr & HAS_BEEN_IDLE) == HAS_BEEN_IDLE) {
                if (state != null) {
                    state += ", ";
                } else {
                    state = "";
                }

                state += "HAS_BEEN_IDLE";
            }

            if (state != null) {
                s += ", state: " + state;
            }

            s += "]";

            return s;
        }

        /**
         *  Writes this component to the given output stream.
         *
         * @param  os               the output stream to write to.
         * @exception  IOException  Description of the Exception
         */
        void write(DataOutputStream os) throws IOException {
            construct();
            super.write(os);
        }

        /**  Description of the Method */
        private void construct() {
            try {
                ByteArrayOutputStream bos = new ByteArrayOutputStream();
                DataOutputStream dos = new DataOutputStream(bos);

                dos.writeShort(sock);
                dos.writeShort(icon);
                dos.writeShort(clr);

                if (nick != null) {
                    // 2002.03.22 MackyBee modified
                    /* Note: String.length depends on the 16 bits Unicode */
                    byte[] nickBytes = nick.getBytes(encoding);
                    dos.writeShort(nickBytes.length);
                    dos.write(nickBytes);
                    //dos.writeShort(nick.length());
                    //dos.write(nick.getBytes(encoding));
                    
                    // 2002.12.23 MackyBee added TEMPORARY. 
                    //System.out.println("[HLProtocol.UserListComponent] " 
                    //    + "nick = " + nick + ", " 
                    //    + "new String(nickBytes, encoding) = "
                    //    + new String(nickBytes, encoding));
                }

                data = bos.toByteArray();

            } catch (IOException e) {}
        }
    }

    /* 2002.05.28 modifies more..
       2002.05.27 MackyBee added experimentally */
    /**  Description of the Class */
    public class NewsBundleListComponent extends DataComponent {

        /* 2002.06.02 MackyBee experimentally added */
        /**  Description of the Field */
        public int componentType;

        /* 2002.06.01 MackyBee disabled. This is not an index value */
        //public int bundleIndex;
        /**  Description of the Field */
        public String bundleTitle;
        /* 2002.05.31 MackyBee added. Note: Actually, this is not a size. */
        /**  Description of the Field */
        public Integer bundleSize;

        /**
         *  Constructor for the NewsBundleListComponent object
         *
         * @param  dis              Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        public NewsBundleListComponent(DataInputStream dis)
             throws IOException {
            super(HTLS_DATA_BUNDLE_LIST, dis);
            DataInputStream din =
                new DataInputStream(new ByteArrayInputStream(data));

            /* 2002.06.02 MackyBee note: It seems that if the component is
               category, the value is 0 3. And modified.
               2002.05.31 MackyBee modified. Temporary.
               Unknown value, it seems to be always 0 2. */
            componentType = din.readUnsignedShort();
            //din.readUnsignedShort();

            /* 2002.06.02 MackyBee provisionally added */
            byte[] titleBytes;

            /* 2002.06.02 MackyBee experimentally added */
            switch (componentType) {
                /* 2002.06.05 MackyBee replaced. */
                case TYPE_BUNDLE:
                    //case 2:
                    /* This is bundle */
                    bundleSize = new Integer(din.readUnsignedShort());
                    titleBytes = new byte[din.readUnsignedByte()];
                    din.read(titleBytes);
                    // 2002.12.28 MackyBee modified. 
                    bundleTitle = new String(titleBytes, encoding);
                    //bundleTitle = new String(titleBytes);
                    break;
                /* 2002.06.05 MackyBee replaced. */
                case TYPE_CATEGORY:
                    //case 3:
                    /* This is category */
                    bundleSize = new Integer(din.readUnsignedShort());
                    /* These data are unknown. */
                    din.skipBytes(24);
                    titleBytes = new byte[din.readUnsignedByte()];
                    din.read(titleBytes);
                    // 2002.12.28 MackyBee modified. 
                    bundleTitle = new String(titleBytes, encoding);
                    //bundleTitle = new String(titleBytes);
                    break;
                default:
                    /* 2002.06.05 MackyBee added. Temporary specification */
                    System.out.println("HLProtocol: " +
                        "NewsComponent type was not fixed.");
                    break;
            }

            /* 2002.06.02 provisionally disabled.
               2002.06.01 MackyBee modified. */
            /*
            bundleSize = new Integer(din.readUnsignedShort());
            //bundleIndex = din.readUnsignedShort();
            byte[] titleBytes = new byte[din.readUnsignedByte()];
            din.read(titleBytes);
            bundleTitle = new String(titleBytes);
            */
        }

        /* 2002.06.02 MackyBee added */
        /**
         *  Constructor for the NewsBundleListComponent object
         *
         * @param  bundleTitleBytes  Description of the Parameter
         * @exception  IOException   Description of the Exception
         */
        public NewsBundleListComponent(byte[] bundleTitleBytes)
             throws IOException {
            super(HLProtocol.HTLC_DATA_NEWS_COMPONENT);

            ByteArrayOutputStream bos = new ByteArrayOutputStream();
            DataOutputStream dos = new DataOutputStream(bos);

            /* Unknown 1 */
            dos.writeShort(1);
            /* Unknown 2 */
            dos.writeShort(0);

            dos.writeByte(bundleTitleBytes.length);
            dos.write(bundleTitleBytes);

            /* Some processes here */
            dos.flush();
            data = bos.toByteArray();
        }

    }

    /* ! Unused now. ! 2002.06.01 MackyBee added */
    /*
    public class NewsCategoryListComponent extends DataComponent {
        public NewsCategoryListComponent(DataInputStream dis)
            throws IOException {
            super(HTLS_DATA_BUNDLE_LIST, dis);
        }
    }
    */
    /* 2002.06.03 MackyBee added */
    /**  Description of the Class */
    public class NewsThreadListComponent extends DataComponent {

        /**  Description of the Field */
        public int threadIndex;
        /**  Description of the Field */
        public int numberOfArticles;

        /**  Description of the Field */
        public int[] articleIndexes;
        /**  Description of the Field */
        public int[] yearfour;

        /* 2002.6.17 MackyBee modified. */
        /**  Description of the Field */
        public long[] secondsOfDays;
        //public int[] dayOfSeconds;
        /**  Description of the Field */
        public int[] repliesFor;

        /**  Description of the Field */
        public String[] titles;
        /**  Description of the Field */
        public String[] posters;
        /**  Description of the Field */
        public String[] mimeTypes;

        /**  Description of the Field */
        public int[] sizeOfBytes;

        /**  Description of the Field */
        public int[] unknown;

        /**
         *  Constructor for the NewsThreadListComponent object
         *
         * @param  dis              Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        public NewsThreadListComponent(DataInputStream dis)
             throws IOException {
            super(HTLS_DATA_NEWS_THREAD, dis);

            DataInputStream tdis =
                new DataInputStream(new ByteArrayInputStream(data));
            /* 2002.06.16 MackyBee EXP added. */
            try {
                /* Provisional specification */
                tdis.skipBytes(4);
                /* Note: Yet care if this is under 0. */
                numberOfArticles = tdis.readInt();

                articleIndexes = new int[numberOfArticles];
                yearfour = new int[numberOfArticles];
                /* 2002.06.17 MackyBee modified. */
                secondsOfDays = new long[numberOfArticles];
                //dayOfSeconds = new int[numberOfArticles];
                repliesFor = new int[numberOfArticles];
                titles = new String[numberOfArticles];
                posters = new String[numberOfArticles];
                mimeTypes = new String[numberOfArticles];
                sizeOfBytes = new int[numberOfArticles];
                unknown = new int[numberOfArticles];

                tdis.skipBytes(2);

                byte[] tempBytes = null;
                long y4 = 0L;
                long y3 = 0L;
                long y2 = 0L;
                long y1 = 0L;
                /* 2002.06.20 MackyBee added a boolean variable. */
                boolean illegalSwitch = false;
                for (int a = 0; a < numberOfArticles; a++) {

                    /* 2002.06.20 MackyBee added. */
                    if (illegalSwitch) {
                        illegalSwitch = false;
                    } else {
                        /* Correct */
                        tdis.skipBytes(2);
                    }

                    articleIndexes[a] = tdis.readUnsignedShort();

                    yearfour[a] = tdis.readUnsignedShort();
                    /* Provisionally */
                    unknown[a] = tdis.readUnsignedShort();

                    /* 2002.06.17 MackyBee modified. */
                    y4 = (long) tdis.readUnsignedByte();
                    y3 = (long) tdis.readUnsignedByte();
                    y2 = (long) tdis.readUnsignedByte();
                    y1 = (long) tdis.readUnsignedByte();
                    secondsOfDays[a] = y4 * 256L * 256L * 256L +
                        y3 * 256L * 256L + y2 * 256L + y1;
                    //dayOfSeconds[a] = tdis.readInt();

                    repliesFor[a] = tdis.readInt();

                    /* Provisional specification */
                    tdis.skipBytes(4);

                    /* 2002.06.16 MackyBee modified.
                       This value indecates if there is "title" data. */
                    if (tdis.readUnsignedShort() == 1) {
                        /* Correct process */
                        tempBytes = new byte[tdis.readUnsignedByte()];
                        tdis.read(tempBytes);
                        // 2002.12.28 MackyBee modified. 
                        titles[a] = new String(tempBytes, encoding);
                        //titles[a] = new String(tempBytes);
                    } else {
                        /* There is no "title" data. */
                        //illegalSwitch = true;
                        tdis.skipBytes(1);
                    }

                    tempBytes = new byte[tdis.readUnsignedByte()];
                    /* 2002.06.20 MackyBee added this if statement. */
                    if (tempBytes.length == 0) {
                        //illegalSwitch = true;
                        posters[a] = "";
                    } else {
                        /* Correct process */
                        tdis.read(tempBytes);
                        // 2002.12.28 MackyBee modified. 
                        posters[a] = new String(tempBytes, encoding);
                        //posters[a] = new String(tempBytes);
                    }

                    tempBytes = new byte[tdis.readUnsignedByte()];
                    /* 2002.06.20 MackyBee added this if statement. */
                    if (tempBytes.length == 0) {
                        illegalSwitch = true;
                        mimeTypes[a] = "";
                    } else {
                        /* Correct process */
                        tdis.read(tempBytes);
                        // 2002.12.28 MackyBee modified. 
                        mimeTypes[a] = new String(tempBytes, encoding);
                        //mimeTypes[a] = new String(tempBytes);
                    }

                    /* 2002.06.20 MackyBee added this if statement.
                       Note: This seems to be a rule of the protocol. */
                    if (illegalSwitch) {
                        /* PROVISIONALLY, input a value */
                        sizeOfBytes[a] = 0;
                    } else {
                        /* Correct process */
                        sizeOfBytes[a] = tdis.readUnsignedShort();
                    }
                }
            } catch (EOFException eofe) {
                DebuggerOutput.debug("NewsThreadListComponent: " +
                    "EOFException occurred.");
            }
        }

        /* Disabled. 2002.06.04 MackyBee added */
        /*
        public NewsThreadListComponent(byte[] bundle, byte[] category)
            throws IOException {
            super(HLProtocol.HTLC_DATA_NEWS_COMPONENT);

            ByteArrayOutputStream bos = new ByteArrayOutputStream();
            DataOutputStream dos = new DataOutputStream(bos);
            */
        /* 2002.06.06 MackyBee added note:
               If specifying category has no parent bundle, this value may
               be {0 1}.
               In case getting news thread, this becomes {0 2}. */
        /*if (bundle == null) {
                dos.writeShort(1);
            } else {
                dos.writeShort(2);*/
        /* Unknown */
        /*dos.writeShort(0);
                dos.writeByte(bundle.length);
                dos.write(bundle);
            }*/
        /* Unknown */
        /*dos.writeShort(0);
            dos.writeByte(category.length);
            dos.write(category);

            dos.flush();
            data = bos.toByteArray();
        }*/
        /* 2002.06.15 MackyBee note: This argument is not needed as
           NewsComponent type.
           2002.06.10 MackyBee added experimentally */
        /**
         *  Constructor for the NewsThreadListComponent object
         *
         * @param  ncs              Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
//        public NewsThreadListComponent(NewsComponent[] ncs)
//             throws IOException {
//            super(HLProtocol.HTLC_DATA_NEWS_COMPONENT);
//
//            /* 2002.06.15 MackyBee added */
//            if (ncs == null) {
//                return;
//            }
//
//            ByteArrayOutputStream bos = new ByteArrayOutputStream();
//            DataOutputStream dos = new DataOutputStream(bos);
//
//            byte[] b = null;
//
//            /* 2002.06.10 MackyBee modified experimentally */
//            dos.writeShort(ncs.length);
//
//            for (int i = 0; i < ncs.length; i++) {
//                /* 2002.06.10 MackyBee modified experimentally */
//                //dos.writeShort(ncs.length);
//                //dos.writeShort(convertTypeValue(ncs[i].getComponentType()));
//                dos.writeShort(0);
//                b = ncs[i].getUserObject().toString().getBytes(encoding);
//                dos.writeByte(b.length);
//                dos.write(b);
//            }
//
//            dos.flush();
//            data = bos.toByteArray();
//        }
        /*
        private int convertTypeValue(int value) {
            return (value - 1);
        }
        */
    }

    /* Unused.
       2002.06.07 MackyBee experimentally added. */
    /*
    public class NewsArticleInfo {
        private String article;
        private String mimeType;
        private int articleIndex;
        protected void setArticle(String article) {
            this.article = article;
        }
        public String getArticle() {
            return (article);
        }
    }
    */
    /* 2002.06.07 MackyBee ! note: !
       This class almost equals to DataComponent. Think again later.
       2002.06.04 MackyBee added. */
    /**  Description of the Class */
    public class NewsArticleComponent extends DataComponent {

        /**  Description of the Field */
        protected String article;

        /**
         *  Constructor for the NewsArticleComponent object
         *
         * @param  dis              Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        public NewsArticleComponent(DataInputStream dis)
             throws IOException {
            super(HTLS_DATA_NEWS_ARTICLE, dis);

            // 2002.12.28 MackyBee modified. 
            // 2002.06.07 MackyBee modified */
            article = new String(data, encoding);
            //article = new String(data);

            //DataInputStream tdis =
            //    new DataInputStream(new ByteArrayInputStream(data));

            //byte[] b = new byte[tdis.readUnsignedShort()];
            //tdis.read(b);
            //article = new String(b);
        }
    }

    /* 2002.06.04 MackyBee added.
       I feel this is some waste... */
    /**  Description of the Class */
    public class NewsArticleIndexComponent extends DataComponent {

        /**  Description of the Field */
        protected int articleIndex;

        /**
         *  Constructor for the NewsArticleIndexComponent object
         *
         * @param  index            Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        public NewsArticleIndexComponent(int index) throws IOException {
            super(HTLC_DATA_NEWS_ARTICLE_INDEX);

            ByteArrayOutputStream bos = new ByteArrayOutputStream();
            DataOutputStream dos = new DataOutputStream(bos);

            dos.writeShort(index);

            dos.flush();
            data = bos.toByteArray();
        }

        /**
         *  Constructor for the NewsArticleIndexComponent object
         *
         * @param  dis              Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        public NewsArticleIndexComponent(DataInputStream dis)
             throws IOException {
            super(HTLC_DATA_NEWS_ARTICLE_INDEX, dis);
            DataInputStream tdis =
                new DataInputStream(new ByteArrayInputStream(data));
            articleIndex = tdis.readUnsignedShort();
        }
    }

    /* 2002.06.04 MackyBee added. */
    /**  Description of the Class */
    public class NewsArticleMIMETypeComponent extends DataComponent {

        /**  Description of the Field */
        public String mimeType;

        /**
         *  Constructor for the NewsArticleMIMETypeComponent object
         *
         * @param  mimeType         Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        public NewsArticleMIMETypeComponent(String mimeType) throws
            IOException {
            super(HTLC_DATA_NEWS_ARTICLE_MIME_TYPE);

            ByteArrayOutputStream bos = new ByteArrayOutputStream();
            DataOutputStream dos = new DataOutputStream(bos);

            dos.write(mimeType.getBytes(encoding));

            dos.flush();
            data = bos.toByteArray();
        }

        /* 2002.06.05 MackyBee added */
        /**
         *  Constructor for the NewsArticleMIMETypeComponent object
         *
         * @param  dis              Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        public NewsArticleMIMETypeComponent(DataInputStream dis)
             throws IOException {
            super(HTLC_DATA_NEWS_ARTICLE_MIME_TYPE, dis);
            // 2002.12.28 MackyBee modified. 
            // 2002.06.07 MackyBee modified
            mimeType = new String(data, encoding);
            //mimeType = new String(data);
            //DataInputStream tdis =
            //    new DataInputStream(new ByteArrayInputStream(data));

            //byte[] b = new byte[tdis.readUnsignedShort()];
            //mimeType = new String(b);
        }
    }

    /**
     *  Most Hotline packets will contain one or more of these. They are
     *  essentially typed data packets, consisting of a type, a length, and the
     *  data itself.<p>
     *
     *  For some types of data there are specialized subclasses of this class,
     *  e.g. {@link HLProtocol.UserListComponent} and {@link DateComponent}
     *  providing an easier interface to the data then a byte array.
     */
    class DataComponent implements Cloneable {

        /**  The data in this data component. */
        public byte[] data;

        /**  The type of this data component. */
        public short type;

        /**
         *  Constructs a DataComponent with specified type and no contents.
         *
         * @param  t  the type of this DataComponent.
         */
        DataComponent(short t) {
            type = t;
            data = new byte[0];
        }

        /**
         *  Constructs a DataComponent with specified type and contents.
         *
         * @param  t  the type of this DataComponent.
         * @param  d  the data in this DataComponent.
         */
        DataComponent(short t, byte[] d) {
            type = t;
            data = (d != null ? d : new byte[0]);
        }

        /**
         *  Constructs a DataComponent of the given type from the given
         *  DataInputStream. The input stream should point at the beginning of
         *  the 16-bit size value for this component, ie. skipping the 16-bit
         *  type value.
         *
         * @param  t                the type of this DataComponent.
         * @param  din              DataInputStream to read from.
         * @exception  IOException  Description of the Exception
         */
        DataComponent(short t, DataInputStream din) throws IOException {

            type = t;

            /* 2002.04.14 MackyBee modified */
            data = new byte[din.readUnsignedShort()];
            /* 2002.03.23 MackyBee modified */
            //data = new byte[(int) din.readShort()];
            //data = new byte[(int) din.readChar()];
            din.readFully(data, 0, data.length);
        }

        /**
         *  Constructs by reading from a DataInputStream.
         *
         * @param  din              DataInputStream to read the hx_data_hdr
         *      from.
         * @exception  IOException  Description of the Exception
         */
        DataComponent(DataInputStream din) throws IOException {
            this(din.readShort(), din);
        }

        /**
         *  Description of the Method
         *
         * @return    Description of the Return Value
         */
        public String toString() {

            String dataString = new String("");

            switch (type) {

                case HTLS_DATA_FILE_SIZE:
                case HTLS_DATA_HTXF_SIZE:
                case HTLS_DATA_OPTION:
                case HTLS_DATA_ICON:
                case HTLS_DATA_COLOUR:
                case HTLS_DATA_SOCKET:
                case HTLS_DATA_VERSION:
                    dataString = new Integer(
                        ToArrayConverters.byteArrayToInt(data)).toString();
                    break;
                case HTLS_DATA_FILE_DATE_CREATED:
                case HTLS_DATA_FILE_DATE_MODIFIED:
                    try {
                        ByteArrayInputStream bis =
                            new ByteArrayInputStream(data);
                        DataInputStream dis = new DataInputStream(bis);
                        dataString = new DateComponent(type, dis).toString();
                    } catch (IOException e) {}
                    break;
                case HTLS_DATA_CHAT_REF:
                case HTLS_DATA_HTXF_REF:
                    dataString = "0x" + Integer.toHexString(
                        ToArrayConverters.byteArrayToInt(data));
                    break;
                default:
                    dataString = new String(data);

                    dataString = TextUtils.findAndReplace(
                        dataString, "\r", "\\r");
                    dataString = TextUtils.findAndReplace(
                        dataString, "\n", "\\n");

                    /* 2002.03.28 MackyBee modified */
                    /* 2002.04.04 MackyBee re-modified */
                    //int restLen = dataString.getBytes(encoding).length;
                    int restLen = dataString.length();
                    if (restLen > 40) {
                        //if (dataString.length() > 40) {
                        //int restLen = dataString.length();
                        dataString = dataString.substring(0, 40);
                        dataString += "... (" + restLen + ")";
                    }
                    break;
            }

            return "DataComponent[type = 0x" + Integer.toHexString(type) +
                ", data = " + dataString + "]";
        }

        /**
         *  Writes this DataComponent to a DataOutputStream.
         *
         * @param  dos              DataOutputStream to write hx_data_hdr to.
         * @exception  IOException  Description of the Exception
         */
        void write(DataOutputStream dos) throws IOException {
            dos.writeChar(type);
            dos.writeChar((char) data.length);
            dos.write(data, 0, data.length);
            dos.flush();
        }

    }

    /**  Encodes a Hotline date. */
    class DateComponent extends DataComponent {

        Calendar date;

        /**
         * @param  type  one of HTLS_DATA_FILE_DATE_CREATED or
         *      HTLS_DATA_FILE_DATE_MODIFIED.
         * @param  when  the date to encode.
         */
        public DateComponent(short type, Calendar when) {
            super(type);
            date = when;
            data = HLProtocol.calendarToHotlineDate(date);
        }

        /**
         * @param  type             one of HTLS_DATA_FILE_DATE_CREATED or
         *      HTLS_DATA_FILE_DATE_MODIFIED.
         * @param  in               the DataInputStream to construct this object
         *      from.
         * @exception  IOException  Description of the Exception
         */
        public DateComponent(short type, DataInputStream in) throws IOException {
            super(type, in);
            date = HLProtocol.hotlineDateToCalendar(data);
        }

        /**
         *  Description of the Method
         *
         * @return    Description of the Return Value
         */
        public String toString() {
            return "DateComponent[type = 0x" + Integer.toHexString(type) + ", " + dateFormatter.format(date.getTime()) + "]";
        }
    }

    /**
     *  Describes a file. {@see ComponentFactory#createPathComponents}.
     */
    class FileComponent extends DataComponent {

        String file;

        /**
         *  Constructor for the FileComponent object
         *
         * @param  file  Description of the Parameter
         */
        public FileComponent(String file) {
            super(HTLC_DATA_FILE);
            this.file = file;
            construct();
        }

        /**
         *  Constructor for the FileComponent object
         *
         * @param  din              Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        public FileComponent(DataInputStream din) throws IOException {
            super(HTLC_DATA_FILE, din);
            // 2002.12.28 MackyBee took NOTE: Needs a consideration. 
            // 2002.11.13 MackyBee modified
            this.file = EncodingUtil.strictGetString(data);
            //this.file = new String(data);
        }

        /**  Description of the Method */
        private void construct() {
            // 2002.12.17 MackyBee modified. 
            // 2002.11.11. MackyBee re-modified.
            // 2002.05.02 MackyBee modified.
            //data = EncodingUtil.strictGetBytes(file);
            try {
                data = file.getBytes(encoding);
            } catch (Exception e) {
                e.printStackTrace();
            }

            // 2002.10.29 MackyBee added TEMPORARY for debug.
            //System.out.println("FileComponent.construct(): "
            //    + "new String(data) = " + new String(data));
        }
    }

    /**
     *  The header which precedes a file transfer, for handshaking purposes
     *  among other things. Is followed by a FileTransferInfo block.
     */
    class FileTransferHeader {

        /**  The length of this transfer header. */
        public long len;

        /**  The ID for this transfer. */
        public int ref;

        /**  Unknown. */
        public int unknown = 0;

        /**
         *  Constructs a transfer header with the specified transfer ID and
         *  header length.
         *
         * @param  r  the transfer ID.
         * @param  l  the transfer header length.
         */
        FileTransferHeader(int r, long l) {
            ref = r;
            len = l;
        }

        /**
         *  Constructor for the FileTransferHeader object
         *
         * @param  din              Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        FileTransferHeader(DataInputStream din) throws IOException {
            read(din);
        }

        /**
         *  Description of the Method
         *
         * @return    Description of the Return Value
         */
        public String toString() {
            return "FileTransferHeader[ref = 0x" + Integer.toHexString(ref)
                 + ", len = " + len + ", unknown = " + unknown + "]";
        }

        /**
         *  Description of the Method
         *
         * @param  din              Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        void read(DataInputStream din) throws IOException {

            int magic = din.readInt();

            if (magic != HLProtocol.HTXF_MAGIC_INT) {
                throw new IOException("not a FileTransferHeader");
            }

            ref = din.readInt();
            int low = din.readInt();
            int high = din.readInt();
            len = (long) high << 32 | (long) low;
        }

        /**
         *  Writes the transfer header to the given DataOutputStream.
         *
         * @param  dos              the DataOutputStream to write to.
         * @exception  IOException  Description of the Exception
         */
        void write(DataOutputStream dos) throws IOException {
            dos.writeInt(HLProtocol.HTXF_MAGIC_INT);
            dos.writeInt(ref);

            /* This is for 63 bit file size support. Don't know
               if it clobbers anything important. */
            int low = (int) (len & 0xffffffffL);
            int high = (int) (len >> 32);
            dos.writeInt(low);
            dos.writeInt(high);
            dos.flush();
        }
    }

    /**
     *  This block describes some meta-data about the file being transferred.
     *  Quite a bit of data is ignored / assumed zero when reading / writing
     *  this info packet, because I don't know it's purpose.
     */
    class FileTransferInfo {

        Calendar creationDate, modificationDate;
        byte[] data;

        String fileName, fileType, fileCreator, fileComment;
        int finderFlags;
        int numberOfBlocks;

        /**
         *  Constructor for the FileTransferInfo object
         *
         * @param  fName            Description of the Parameter
         * @param  fType            Description of the Parameter
         * @param  fCreator         Description of the Parameter
         * @param  fComment         Description of the Parameter
         * @param  created          Description of the Parameter
         * @param  modified         Description of the Parameter
         * @param  flags            Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        public FileTransferInfo(String fName,
            String fType,
            String fCreator,
            String fComment,
            Calendar created,
            Calendar modified,
            int flags) throws IOException {

            /* 2002.03.28 MackyBee modified. */
            if (fComment.getBytes(encoding).length > 255) {
                //if (fComment.length() > 255) {
                throw new IllegalArgumentException(
                    "comment length must be <= 255.");
            }

            if (fType.length() != 4) {
                throw new IllegalArgumentException(
                    "file type must be 4 bytes.");
            }

            if (fCreator.length() != 4) {
                throw new IllegalArgumentException(
                    "file creator must be 4 bytes.");
            }

            fileName = fName;
            fileType = fType;
            fileCreator = fCreator;
            fileComment = fComment;
            creationDate = created;
            modificationDate = modified;
            finderFlags = flags;

            construct();
        }

        /**
         *  Constructor for the FileTransferInfo object
         *
         * @param  originalInput    Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        public FileTransferInfo(DataInputStream originalInput)
             throws IOException {

            DataInputStream input = originalInput;
            ByteArrayOutputStream monitorBos = new ByteArrayOutputStream();
            DataOutputStream monitorDos = new DataOutputStream(monitorBos);

            byte[] filp = new byte[40];
            originalInput.readFully(filp, 0, 40);
            monitorDos.write(filp);
            ByteArrayInputStream bis = new ByteArrayInputStream(filp);
            input = new DataInputStream(bis);

            /* Get the FILP magic. */
            byte[] magic = new byte[4];
            input.readFully(magic, 0, 4);

            if (!new String(magic).equals("FILP")) {
                throw new IOException("Not a FileTransferInfo - bad magic.");
            }

            /* Get some unknown parameter. */
            int parameter = input.readShort();

            if (parameter != 1) {
                throw new IOException("Scared of unknown parameter 1 (is "
                     + parameter + ", expected 1).");
            }

            /* Skip some unknown stuff. */
            input.skipBytes(14);

            /* Get another unknown parameter. Number of blocks??
               I think this is the number of blocks
               that are being sent (e.g. 2 for FILP + DATA or 3 for
               FILP + DATA + MACR). */
            numberOfBlocks = input.readInt();

            if (!(numberOfBlocks == 3 || numberOfBlocks == 2)) {
                throw new IOException("Scared of numberOfBlocks value (is " + numberOfBlocks + ", expected 2 or 3).");
            }

            /* Get the INFO magic. */
            input.readFully(magic, 0, 4);

            if (!new String(magic).equals("INFO")) {
                throw new IOException("Got " + new String(magic)
                     + ", expected INFO.");
            }

            /* Skip some unknown stuff. */
            input.skipBytes(8);

            /* Get the remaining header size. */
            int headerRemaining = input.readInt();

            if (headerRemaining > 64 * 1024) {
                throw new IOException(
                    "Scared of too large FileTransferInfo size ("
                     + headerRemaining + ").");
            }

            /* Read the remaining header, up until the DATA block. */
            byte[] amac = new byte[headerRemaining];
            originalInput.readFully(amac, 0, headerRemaining);
            monitorDos.write(amac);
            monitorDos.flush();
            data = monitorBos.toByteArray();

            if (hexdumpPayload) {
                DebuggerOutput.debug("Dumping incoming file transfer info:");
                DebuggerOutput.dumpBytes(data, 8);
            }

            bis = new ByteArrayInputStream(amac);
            input = new DataInputStream(bis);

            /* Get the AMAC / MWIN magic. */
            input.readFully(magic, 0, 4);

            if (!(new String(magic).equals("AMAC") ||
                new String(magic).equals("MWIN"))) {
                throw new IOException("Got " + new String(magic)
                     + ", expected AMAC or MWIN.");
            }

            /* Get the type and creator info. */
            input.readFully(magic, 0, 4);
            fileType = new String(magic);
            input.readFully(magic, 0, 4);
            fileCreator = new String(magic);

            /* Skip some unknown stuff. */
            input.skipBytes(6);

            /* Skip an unknown parameter (perhaps Finder flags?
               21 for an SimpleText, 0 and 1 for some other files).*/
            finderFlags = input.readUnsignedByte();

            /* Skip more unknown stuff. */
            input.skipBytes(33);

            /* Get the creation and modification date. */
            byte[] date = new byte[8];
            input.readFully(date, 0, 8);
            creationDate = HLProtocol.hotlineDateToCalendar(date);
            input.readFully(date, 0, 8);
            modificationDate = HLProtocol.hotlineDateToCalendar(date);

            /* Get the file name. */
            headerRemaining -= 60;
            int filenameLength = input.readInt();

            if (filenameLength > headerRemaining) {
                throw new IOException(
                    "FileTransferInfo is inconsistent, "
                     + "filename exceeds total header length.");
            }

            byte[] fName = new byte[filenameLength];
            input.readFully(fName, 0, filenameLength);
            // 2002.11.11 MackyBee experimentally modified.
            fileName = EncodingUtil.strictGetString(fName);
            //fileName = new String(fName);
            input.skipBytes(1);

            // 2002.11.12 TEMPORARY
            //System.out.println("HLProtocol [fileName]: " + fileName);
            // 2002.11.12 TEMPORARY
            //System.out.println("HLProtocol [fName]: ");
            //for (int fn = 0; fn < fName.length; fn++) {
            //    System.out.print(String.valueOf(fName[fn]) + " ");
            //}
            //System.out.println();
            //System.out.println("--");

            /* Get the comment. */
            headerRemaining -= filenameLength + 2;
            /* 2002.04.14 MackyBee modified */
            char commentLength = (char) input.readUnsignedByte();
            //char commentLength = (char) SignUtils.toUnsigned(input.readByte());

            if (commentLength > headerRemaining) {
                throw new IOException("FileTransferInfo is inconsistent, "
                     + "comment length (" + (int) commentLength
                     + ") exceeds total header length (" + headerRemaining
                     + ").");
            }

            byte[] fComment = new byte[commentLength];
            input.readFully(fComment, 0, commentLength);
            fileComment = new String(fComment);

            input.skipBytes(1);
        }

        /**
         *  Description of the Method
         *
         * @param  output           Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        public void write(DataOutputStream output) throws IOException {
            output.write(data);
            output.flush();
        }

        /**
         *  Description of the Method
         *
         * @exception  IOException  Description of the Exception
         */
        private void construct() throws IOException {

            ByteArrayOutputStream monitorBos = new ByteArrayOutputStream();
            DataOutputStream output = new DataOutputStream(monitorBos);

            byte[] buf = new byte[33];

            output.write(new String("FILP").getBytes(encoding));
            output.writeShort(1);
            output.write(buf, 0, 14);

            // 2002.07.26 MackyBee modified.
            // Note: According to official client protocol, here is 2.
            // But this may changes case by case...
            output.writeInt(2);
            //output.writeInt(3);

            output.write(new String("INFO").getBytes(encoding));
            output.write(buf, 0, 8);

            ByteArrayOutputStream bos = new ByteArrayOutputStream();
            DataOutputStream dos = new DataOutputStream(bos);

            // 2002.07.26 MackyBee EXPERIMENTALLY replaced.
            dos.write(new String("MWIN").getBytes(encoding));
            //dos.write(new String("AMAC").getBytes(encoding));

            dos.write(fileType.getBytes(encoding));
            dos.write(fileCreator.getBytes(encoding));
            dos.write(buf, 0, 6);
            dos.writeByte((byte) finderFlags);
            dos.write(buf, 0, 33);

            // 2002.07.26 MackyBee modified.
            // Note: According to an analyzed official client protocol record.
            dos.write(buf, 0, 16);
            /*
            byte[] created = HLProtocol.calendarToHotlineDate(creationDate);
            byte[] modified = HLProtocol.calendarToHotlineDate(modificationDate);
            dos.write(created);
            dos.write(modified);
            */
            // 2002.12.17 MackyBee modified. 
            // 2002.11.11 MackyBee modified.
            // 2002.03.28 MackyBee modified
            //byte[] fnBytes = EncodingUtil.strictGetBytes(fileName);
            byte[] fnBytes = fileName.getBytes(encoding);
            dos.writeInt(fnBytes.length);
            dos.write(fnBytes);
            //dos.writeInt(fileName.length());
            //dos.write(fileName.getBytes(encoding));
            dos.writeByte(0);

            // 2002.12.17 MackyBee modified. 
            // 2002.11.11 MackyBee modified.
            // 2002.03.28 MackyBee modified
            //byte[] fcBytes = EncodingUtil.strictGetBytes(fileComment);
            byte[] fcBytes = fileComment.getBytes(encoding);
            dos.writeByte((byte) fcBytes.length);
            dos.write(fcBytes);
            //dos.writeByte((short) fileComment.length());
            //dos.write(fileComment.getBytes(encoding));

            // 2002.07.26 MackyBee disabled.
            //dos.writeByte(0);

            dos.flush();

            output.writeInt(bos.toByteArray().length);
            output.write(bos.toByteArray());
            output.flush();

            if (hexdumpPayload) {
                DebuggerOutput.debug("Dumping outgoing file transfer header:");
                DebuggerOutput.dumpBytes(monitorBos.toByteArray(), 8);
            }

            data = monitorBos.toByteArray();
        }
    }

    /**
     *  Objectified Hotline packet. A Hotline packet consists of:<p>
     *
     *  - A {@link HLProtocol.PacketHeader}.<br>
     *  - Zero or more {@link HLProtocol.DataComponent}'s.<br>
     *  The packet header is accessible via the 'header' member. DataComponents
     *  are accessible via the 'hasMoreComponents' and 'nextComponent' methods
     *  of the Packet class.
     */
    class Packet {

        /* Zero or more DataComponents. */
        DataComponent[] dataComponents;

        /* The header of this packet. */
        PacketHeader header;

        /**
         *  Reads a Hotline packet from the given DataInputStream and
         *  initializes the header and dataComponents fields.
         *
         * @param  input            the DataInputStream to read from.
         * @exception  IOException  Description of the Exception
         */
        Packet(DataInputStream input) throws IOException {

            /* Added experimentally 2002.02.15 */
            //DebuggerOutput.debug(
            //	"HLProtocol.Packet: Getting a PacketHeader...");

            header = new PacketHeader(input);
            /* Added experimentally 2002.03.02 */
            //DebuggerOutput.debug(
            //	"HLProtocol.Packet: Got the PacketHeader...");

            /* Added experimentally 2002.02.15 */
            //DebuggerOutput.debug(
            //"HLProtocol.Packet [header.hc] = " + header.hc);

            if (header.len >= 2) {

                byte[] payload = new byte[(int) header.len - 2];

                /* Added experimentally 2002.02.15 by MackyBee */
                //DebuggerOutput.debug(
                //	"HLProtocol.Packet: reading payload...");

                input.readFully(payload, 0, payload.length);

                if (hexdumpPayload) {
                    DebuggerOutput.debug("Dumping payload for incoming packet:");
                    DebuggerOutput.dumpBytes(payload, 8);
                }

                ByteArrayInputStream bis = new ByteArrayInputStream(payload);
                DataInputStream dis = new DataInputStream(bis);

                /* The following code incorporates a workaround for a
                   Tube 1.0b2 bug. The problem is that Tube 1.0b2
                   sometimes sends packets where the header.hc field
                   does not match the actual number of components. So
                   we simply read until EOF and allow the creation of
                   null components, then prune that stuff afterwards. */
                /* 2002.05.01 MackyBee experimentally modified */
                dataComponents = new DataComponent[header.hc];
                for (int i = 0; i < header.hc; i++) {

                    //DataComponent[] specifiedComponents =
                    //    new DataComponent[header.hc];

                    //short actualComponentCount = 0;

                    //for (int i = 0; i < specifiedComponents.length; i++) {

                    //specifiedComponents[i] =
                    // 2002.12.22 MackyBee modified. 
                    dataComponents[i] = createComponent(dis);
                    //dataComponents[i] =
                        //ComponentFactory.createComponent(dis);

                    //    if (specifiedComponents[i] != null) {
                    //        actualComponentCount++;
                    //    }
                }

                //dataComponents = new DataComponent[actualComponentCount];

                //int j = 0;

                //for (int i = 0; i < specifiedComponents.length; i++) {
                //    if (specifiedComponents[i] != null) {
                //        dataComponents[j++] = specifiedComponents[i];
                //    }
                //}

                //header.hc = actualComponentCount;

            }
        }

        /**
         *  Creates a Packet with the given values for later writing with the
         *  write() method.
         *
         * @param  type            the type of this packet.
         * @param  trans           the transaction number of this packet.
         * @param  dataComponents  the data components in this packet. null is
         *      allowed.
         */
        // 2002.11.21 MackyBee modified a little in the method description.
        Packet(int type, int trans, DataComponent[] dataComponents) {

            if (dataComponents == null) {
                dataComponents = new DataComponent[0];
            }

            header = new PacketHeader(0, type, trans, 0, 2 +
                (SIZEOF_DATACOMPONENT * dataComponents.length), 0);

            /* Compute the total length of this packet. */
            for (int k = 0; k < dataComponents.length; k++) {
                if (dataComponents[k].data != null) {
                    header.len += dataComponents[k].data.length;
                }
            }

            header.len2 = header.len;
            /* 2002.04.14 MackyBee modified */
            header.hc = dataComponents.length;
            //header.hc = (short) dataComponents.length;

            this.dataComponents = dataComponents;
        }

        /**
         *  Returns a String representation of this packet (for debugging
         *  purposes). Slow.
         *
         * @return    String.
         */
        public String toString() {
            if (!DebuggerOutput.on) {
                return "Packet[debugging output disabled]";
            }

            String s = "Packet[" + header.toString() + ", ";

            if (dataComponents == null) {
                s += "null]";
            } else {
                s += "[";
                for (int i = 0; i < dataComponents.length; i++) {
                    if (dataComponents[i] != null) {
                        s += dataComponents[i].toString();
                        if (i + 1 < dataComponents.length) {
                            s += ", ";
                        }
                    }
                }
                s += "]";
            }
            s += "]";
            return s;
        }

        /**
         *  Returns true while there are more DataComponent's in this packet.
         *
         * @return    Description of the Return Value
         */
        boolean hasMoreComponents() {
            return (header.hc > 0 ? true : false);
        }

        /**
         *  Returns the next DataComponent for this packet.
         *
         * @return    Description of the Return Value
         */
        DataComponent nextComponent() {
            return dataComponents[--header.hc];
        }

        /**
         *  Writes the packet to the given DataOutputStream.
         *
         * @param  output           the DataOutputStream to write to.
         * @exception  IOException  Description of the Exception
         */
        void write(DataOutputStream output) throws IOException {

            ByteArrayOutputStream bos = new ByteArrayOutputStream();
            DataOutputStream dos = new DataOutputStream(bos);

            for (int k = 0; k < dataComponents.length; k++) {
                dataComponents[k].write(dos);
            }

            dos.flush();

            byte[] payload = bos.toByteArray();

            if (hexdumpPayload) {
                DebuggerOutput.debug("Dumping payload for outgoing packet:");
                DebuggerOutput.dumpBytes(payload, 8);
            }

            synchronized (output) {
                header.write(output);
                output.write(payload);
                output.flush();
            }

            // 2002.09.05 MackyBee note:
            // I'd like to move this process to caller class for reducing
            // client function specific process.
            /* Added 2002.03.08 MackyBee */
            /* 2002.03.15 re-changed, 2002.03.21 modified */
            if (pps != null) {
                if (header.id != HTLC_HDR_USER_LIST) {
                    /* 	Counts idling time until next sending */
                    pps.setWait(180000L);
                }
            }
        }

    }

    /**
     *  The header of a Hotline packet, describing the length and the kind of
     *  packet among other things.
     */
    class PacketHeader {

        /**  Whether this is a reply to a request or not. */
        public int cls;

        /* 2002.04.14 MackyBee modified experimentally.
           Since the "hc" has to be a real value (a actual number) and it will
           be able to used by readUnsignedShort().
           When writing "hc" value to a stream, it has to be a short value, but
           "down casting" is not a matter. */
        /**  Number of data components in this packet. */
        public int hc;
        //public short hc;

        /**  Kind of packet. */
        public int id;

        /**  Error flag. */
        public int isError;

        /**  Length of packet. */
        public long len;

        /**  Length of packet. */
        public long len2;

        /**  Transaction ID. */
        public int trans;


        /**  Constructs empty. */
        public PacketHeader() { }


        /**
         *  Constructs with specified values.
         *
         * @param  t        cls
         * @param  i        type of packet
         * @param  tr       transaction ID of packet
         * @param  isError  error flag
         * @param  l        length of packet
         * @param  c        Description of the Parameter
         */
        public PacketHeader(int t, int i, int tr, int isError, int l, int c) {

            cls = t;
            id = i;
            trans = tr;
            this.isError = isError;
            len = len2 = l;
            /* 2002.04.14 MackyBee modified */
            hc = c;
            //hc = (short) c;

        }


        /**
         *  Constructs from specified DataInputStream.
         *
         * @param  din              DataInputStream to read from.
         * @exception  IOException  Description of the Exception
         */
        public PacketHeader(DataInputStream din) throws IOException {

            read(din);

        }


        /**
         *  Reads from a DataInputStream.
         *
         * @param  din           DataInputStream to read from.
         * @throws  IOException  if the packet does not appear to be a Hotline
         *      packet.
         */
        public void read(DataInputStream din) throws IOException {

            if (din == null) {
                DebuggerOutput.debug("DataInputStream called with NULL");
            }

            /* 2002.02.15 MackyBee added. 2002.03.23 modified. */
            //DebuggerOutput.debug("PacketHeader: read[id]");

            int temp = din.readInt();

            cls = temp >> 16;
            id = (temp == HTLS_HDR_TASK ? HTLS_HDR_TASK : (temp & 0xffff));

            /* Added experimentally 2002.03.03 by MackyBee */
            //DebuggerOutput.debug("PacketHeader.read [id] = " + id);

            /* 2002.02.15 MackyBee added. 2002.03.23 modified. */
            //DebuggerOutput.debug("PacketHeader: read[trans]");

            trans = din.readInt();

            /* Added experimentally 2002.02.15 by MackyBee */
            //DebuggerOutput.debug("PacketHeader: read/isError");

            isError = din.readInt();
            len = din.readInt();
            len2 = din.readInt();

            hc = 0;

            if (len != len2) {
                throw new IOException("Not a Hotline packet");
            }

            if (len >= 2) {
                /* 2002.04.14 MackyBee modified experimentally */
                /* For get real value of length, should use the
                   readUnsignedShort(). But number of data components will not
                   be over 32767 value (if over 32767, it becomes -32768). */
                hc = din.readUnsignedShort();
                //hc = din.readShort();
            }
        }


        /**
         *  Description of the Method
         *
         * @return    Description of the Return Value
         */
        public String toString() {
            return "PacketHeader[cls = 0x" + Integer.toHexString(this.cls) +
                ", id = 0x" + Integer.toHexString(this.id) + ", trans = " +
                this.trans + ", isError = " + this.isError + ", len = " +
                this.len + ", len2 = " + this.len2 + ", hc = " + this.hc + "]";
        }


        /**
         *  Writes to a DataOutputStream.
         *
         * @param  dos              DataOutputStream to write to.
         * @exception  IOException  Description of the Exception
         */
        public void write(DataOutputStream dos) throws IOException {

            int temp = ((int) cls << 16) | id;
            dos.writeInt(temp);
            dos.writeInt(trans);
            dos.writeInt(isError);
            dos.writeInt((int) len);
            dos.writeInt((int) len2);
            /* 2002.04.14 MackyBee modified */
            dos.writeShort((short) hc);
            //dos.writeShort(hc);

            /* 2002.04.29 MackyBee disabled. Because flush is called in
               Packet. */
            //dos.flush();

        }

    }

    /**
     *  Describes a Hotline path. {@see ComponentFactory#createPathComponents}.
     */
    class PathComponent extends DataComponent {

        /**  Description of the Field */
        public String path;

        /**
         *  Constructor for the PathComponent object
         *
         * @param  dataType  Description of the Parameter
         * @param  path      Description of the Parameter
         */
        public PathComponent(short dataType, String path) {
            super(dataType);
            this.path = path;
            construct();
        }

        /**
         *  Interprets a given DataComponent as a PathComponent and initializes
         *  the object with the resulting values.
         *
         * @param  dataType                   Description of the Parameter
         * @param  dis                        Description of the Parameter
         * @exception  IOException            Description of the Exception
         * @throws  IllegalArgumentException  if the supplied input stream does
         *      not point at a ResumeTransferComponent.
         */
        PathComponent(short dataType, DataInputStream dis) throws IOException {

            super(dataType, dis);

            ByteArrayInputStream bis = new ByteArrayInputStream(data);
            DataInputStream is = new DataInputStream(bis);

            /* 2002.04.14 MackyBee modified */
            int dc = is.readUnsignedShort();
            /* 2002.03.23 MackyBee modified experimentally */
            //int dc = (int) is.readShort();
            //int dc = (int) is.readChar();

            /* Extract all the components from the path and use them
               to initialize the 'path' member field. */
            path = "";

            if (dc == 0) {
                path = String.valueOf(DIR_SEPARATOR);
            }

            while (dc-- > 0) {
                short z = is.readShort();
                /* Skip, unknown. */
                byte[] el = new byte[is.readUnsignedByte()];
                is.read(el, 0, el.length);

                // 2002.11.13 MackyBee EXPERIMENTALLY modified.
                path += DIR_SEPARATOR + EncodingUtil.strictGetString(el);
                //path += DIR_SEPARATOR + new String(el);
            }

            /* Recreate the bytestream from the just initialized path
               field. Wasteful, might just use the already given
               bytestream. */
            construct();
        }

        /**
         *  Description of the Method
         *
         * @return    Description of the Return Value
         */
        public String toString() {
            return "PathComponent[type = 0x" + Integer.toHexString(type)
                 + ", " + path + "]";
        }

        /**  Description of the Method */
        private void construct() {
            try {
                ByteArrayOutputStream bos = new ByteArrayOutputStream();
                DataOutputStream dos = new DataOutputStream(bos);

                StringBuffer buf = new StringBuffer(path);
                StringTokenizer st = new StringTokenizer(
                    path, String.valueOf(DIR_SEPARATOR));

                dos.writeChar((char) st.countTokens());

                /* 2002.05.02 MackyBee added and modified.
                   -> disabled a part. */
                byte[] fbytes = null;

                /* Write the path piece by piece. */
                while (st.hasMoreTokens()) {
                    String f = st.nextToken();
                    // 2002.12.17 MackyBee modified. 
                    // 2002.11.11 MackyBee modified.
                    //fbytes = EncodingUtil.strictGetBytes(f);
                    fbytes = f.getBytes(encoding);

                    dos.writeShort(0);
                    dos.writeByte((byte) fbytes.length);
                    //dos.writeByte((byte) f.getBytes(encoding).length);
                    dos.write(fbytes, 0, fbytes.length);
                    //dos.write(f.getBytes(encoding), 0, f.getBytes(encoding).length);
                }

                dos.flush();
                data = bos.toByteArray();

            } catch (IOException e) {}
        }
    }

    /**  Resume file transfer component. */
    class ResumeTransferComponent extends DataComponent {

        /**
         *  The chunks described by this transfer block. (e.g. key = DATA, value
         *  = <size of data chunk>).
         */
        Hashtable chunks;

        /**
         *  Constructs a ResumeTransferComponent with default values.
         */
        ResumeTransferComponent() {
            this(0, 0);
        }

        /**
         *  Constructs a ResumeTransferComponent with the given values.
         *
         * @param  dataSize  Description of the Parameter
         * @param  macrSize  Description of the Parameter
         */
        ResumeTransferComponent(long dataSize, long macrSize) {
            super(HTLS_DATA_HTXF_RFLT);

            chunks = new Hashtable();
            chunks.put("DATA", new Long(dataSize));
            chunks.put("MACR", new Long(macrSize));

            construct();
        }

        /**
         *  Interprets a given DataComponent as a ResumeTransferComponent and
         *  initializes the object with the resulting values.
         *
         * @param  is                         Description of the Parameter
         * @exception  IOException            Description of the Exception
         * @throws  IllegalArgumentException  if the supplied input stream does
         *      not point at a ResumeTransferComponent.
         */
        ResumeTransferComponent(DataInputStream is) throws IOException {

            super(HTLS_DATA_HTXF_RFLT, is);

            chunks = new Hashtable();
            chunks.put("DATA", new Long(0));
            chunks.put("MACR", new Long(0));

            DataInputStream din =
                new DataInputStream(new ByteArrayInputStream(data));

            byte[] id = new byte[4];
            din.readFully(id, 0, 4);

            if (new String(id).equals("RFLT")) {

                byte[] not_interesting = new byte[38];
                din.readFully(not_interesting, 0, 38);

                int available = data.length - (4 + 38);

                while (available > 0) {
                    /* Use 64 bit long for chunk sizes here. This is a
                       protocol extension (or I haven't seen it
                       yet). */
                    din.readFully(id, 0, 4);
                    int low = din.readInt();
                    int high = din.readInt();
                    long chunkSize = (long) high << 32 | (long) low;
                    din.readInt();
                    chunks.put(new String(id), new Long(chunkSize));
                    available -= 16;
                }

                construct();

            } else {
                throw new IllegalArgumentException(
                    "not a ResumeTransferComponent");
            }
        }

        /**
         *  Description of the Method
         *
         * @return    Description of the Return Value
         */
        public String toString() {
            return "ResumeTransferComponent[DATA = " +
                (Long) chunks.get("DATA") + ", MACR = " +
                (Long) chunks.get("MACR") +
                "]";
        }

        /**  Description of the Method */
        private void construct() {

            try {
                ByteArrayOutputStream bos = new ByteArrayOutputStream();
                DataOutputStream dos = new DataOutputStream(bos);

                dos.write(new String("RFLT").getBytes(encoding));
                dos.writeShort(1);
                dos.write(new byte[34], 0, 34);
                dos.writeShort(2);
                dos.write(new String("DATA").getBytes(encoding));
                long ds = ((Long) chunks.get("DATA")).longValue();
                int low = (int) (ds & 0xffffffffL);
                int high = (int) (ds >> 32);
                dos.writeInt(low);
                dos.writeInt(high);
                dos.writeInt(0);
                dos.write(new String("MACR").getBytes(encoding));
                long rs = ((Long) chunks.get("MACR")).longValue();
                low = (int) (rs & 0xffffffffL);
                high = (int) (rs >> 32);
                dos.writeInt(low);
                dos.writeInt(high);
                dos.writeInt(0);
                dos.flush();

                data = bos.toByteArray();
            } catch (IOException e) {}
        }
    }

    /**  Description of the Class */
    class TrackerPacketHeader {

        /**  Description of the Field */
        public short nservers;
        /**  Description of the Field */
        public short pack;

        /**  Description of the Field */
        public byte[] unknown;

        /**
         *  Constructor for the TrackerPacketHeader object
         *
         * @param  din              Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        public TrackerPacketHeader(DataInputStream din) throws IOException {
            read(din);
        }

        /**
         *  Description of the Method
         *
         * @param  din              Description of the Parameter
         * @exception  IOException  Description of the Exception
         */
        void read(DataInputStream din) throws IOException {
            byte[] unknown = new byte[4];
            din.readFully(unknown, 0, 4);
            nservers = din.readShort();
            pack = din.readShort();
        }
    }
}

//{{{ deactivated 
/**  Factory methods for obtaining and creating DataComponents. */
// 2002.12.22 MackyBee EXPERIMENTALLY deactivated. 
// 2002.12.22 MackyBee took NOTE: 
// Does this class really need to be separated from HLProtocol?
//class ComponentFactory {
//
//    static HLProtocol hlp = new HLProtocol();
//
//    /**
//     * Sets an encoding into inner HLProtocol object.
//     */
      //2002.12.21 MackyBee added. 
//    static void setEncoding(String encoding) {
//        hlp.setEncoding(encoding);
//    }
//    
//    /**
//     *  Constructs a DataComponent from the given DataInputStream. For any given
//     *  DataComponent on the input, this method creates and returns the most
//     *  specific subclass of DataComponent that it knows about.
//     *
//     * @param  din  the input stream to construct from.
//     * @return      most specific subclass of DataComponent that is possible.
//     */
//    static HLProtocol.DataComponent createComponent(DataInputStream din) {
//
//        short type;
//
//        try {
//            /* Read the type of this component, then pass object
//               creation on to the appropriate constructors, or create
//               a generic DataComponent if there is no specialized
//               object to represent the given type. */
//            type = din.readShort();
//
//            switch (type) {
//
//                case HLProtocol.HTLS_DATA_HTXF_RFLT:
//                    return hlp.new ResumeTransferComponent(din);
//                case HLProtocol.HTLS_DATA_FILE_LIST:
//                    return hlp.new FileListComponent(din);
//                case HLProtocol.HTLS_DATA_USER_LIST:
//                    return hlp.new UserListComponent(din);
//                case HLProtocol.HTLC_DATA_DIR:
//                case HLProtocol.HTLC_DATA_DIR_RENAME:
//                    return hlp.new PathComponent(type, din);
//                case HLProtocol.HTLC_DATA_FILE:
//                    return hlp.new FileComponent(din);
//                case HLProtocol.HTLS_DATA_FILE_DATE_CREATED:
//                case HLProtocol.HTLS_DATA_FILE_DATE_MODIFIED:
//                    return hlp.new DateComponent(type, din);
//                /* 2002.06.01 MackyBee note:
//                   There is no ID difference bundle and category.
//                   2002.05.31 MackyBee added */
//                case HLProtocol.HTLS_DATA_BUNDLE_LIST:
//                    return hlp.new NewsBundleListComponent(din);
//                /* 2002.06.04 MackyBee added */
//                case HLProtocol.HTLS_DATA_NEWS_THREAD:
//                    return hlp.new NewsThreadListComponent(din);
//                /* 2002.06.05 MackyBee added */
//                case HLProtocol.HTLC_DATA_NEWS_ARTICLE_MIME_TYPE:
//                    return hlp.new NewsArticleMIMETypeComponent(din);
//                /* 2002.06.07 MackyBee added */
//                case HLProtocol.HTLS_DATA_NEWS_ARTICLE:
//                    return hlp.new NewsArticleComponent(din);
//                case HLProtocol.HTLC_DATA_NEWS_ARTICLE_INDEX:
//                    return hlp.new NewsArticleIndexComponent(din);
//                default:
//                    return hlp.new DataComponent(type, din);
//            }
//
//        } catch (EOFException e) {
//            /* This is to compensate for a bug in Tube 1.0b2. Tube
//               1.0b2 doesn't get the PacketHeader.hc field right. See
//               also the PacketHeader constructor. */
//            return null;
//        } catch (IOException e) {
//            e.printStackTrace();
//        }
//
//        return null;
//    }
//
//    /**
//     *  Returns an array containing one or two DataComponents which describing
//     *  the requested path. The number of DataComponents returns depends on the
//     *  value of isFile: if it is true, 2 components are returned.
//     *
//     * @param  path             the Hotline path.
//     * @param  dataType         {@link HLProtocol.HTLC_DATA_DIR} or {@link
//     *      HLProtocol.HTLC_DATA_DIR_RENAME}.
//     * @param  isFile           true if the path refers to a file.
//     * @return                  Description of the Return Value
//     * @exception  IOException  Description of the Exception
//     */
//    static HLProtocol.DataComponent[] createPathComponents(String path,
//        short dataType,
//        boolean isFile) throws IOException {
//
//        HLProtocol.DataComponent[] dataComponents =
//            new HLProtocol.DataComponent[isFile ? 2 : 1];
//
//        if (isFile) {
//            String file = path.substring(path.lastIndexOf(
//                HLProtocol.DIR_SEPARATOR) + 1);
//            dataComponents[1] = hlp.new FileComponent(file);
//
//            /* 2002.04.04 MackyBee modified -> return to original */
//            path = path.substring(0, path.getBytes(encoding).length -
//                file.getBytes(encoding).length);
//            path = path.substring(0, path.length() - file.length());
//        }
//
//        dataComponents[0] = hlp.new PathComponent(dataType, path);
//
//        return dataComponents;
//    }
//}
//}}}
